@@ -1,5 +1,12 @@
 // Start of tuning.h.
 
+
+int is_blank_line_or_comment(const char *s) {
+  size_t i = strspn(s, " \t\n");
+  return s[i] == '\0' || // Line is blank.
+         strncmp(s + i, "--", 2) == 0; // Line is comment.
+}
+
 static char* load_tuning_file(const char *fname,
                               void *cfg,
                               int (*set_tuning_param)(void*, const char*, size_t)) {
@@ -16,10 +23,19 @@ static char* load_tuning_file(const char *fname,
   int lineno = 0;
   while (fgets(line, max_line_len, f) != NULL) {
     lineno++;
+    if (is_blank_line_or_comment(line)) {
+      continue;
+    }
     char *eql = strstr(line, "=");
     if (eql) {
       *eql = 0;
-      int value = atoi(eql+1);
+      char *endptr;
+      int value = strtol(eql+1, &endptr, 10);
+      if (*endptr && *endptr != '\n') {
+        snprintf(line, max_line_len, "Invalid line %d (must be of form 'name=int').",
+                 lineno);
+        return line;
+      }
       if (set_tuning_param(cfg, line, (size_t)value) != 0) {
         char* err = (char*) malloc(max_line_len + 50);
         snprintf(err, max_line_len + 50, "Unknown name '%s' on line %d.", line, lineno);
