@@ -4,6 +4,10 @@ import numpy as np
 import math
 import struct
 
+pi16 = np.float16(np.pi)
+pi32 = np.float32(np.pi)
+pi64 = np.float64(np.pi)
+
 
 def intlit(t, x):
     if t == np.int8:
@@ -596,6 +600,10 @@ def futhark_sqrt64(x):
     return np.sqrt(x)
 
 
+def futhark_rsqrt64(x):
+    return 1 / np.sqrt(x)
+
+
 def futhark_cbrt64(x):
     return np.cbrt(x)
 
@@ -608,26 +616,50 @@ def futhark_cos64(x):
     return np.cos(x)
 
 
+def futhark_cospi64(x):
+    return np.cos(pi64 * x)
+
+
 def futhark_sin64(x):
     return np.sin(x)
 
 
+def futhark_sinpi64(x):
+    return np.sin(pi64 * x)
+
+
 def futhark_tan64(x):
     return np.tan(x)
 
 
+def futhark_tanpi64(x):
+    return np.tan(pi64 * x)
+
+
 def futhark_acos64(x):
     return np.arccos(x)
 
 
+def futhark_acospi64(x):
+    return np.arccos(x) / pi64
+
+
 def futhark_asin64(x):
     return np.arcsin(x)
 
 
+def futhark_asinpi64(x):
+    return np.arcsin(x) / pi64
+
+
 def futhark_atan64(x):
     return np.arctan(x)
 
 
+def futhark_atanpi64(x):
+    return np.arctan(x) / pi64
+
+
 def futhark_cosh64(x):
     return np.cosh(x)
 
@@ -656,6 +688,10 @@ def futhark_atan2_64(x, y):
     return np.arctan2(x, y)
 
 
+def futhark_atan2pi_64(x, y):
+    return np.arctan2(x, y) / pi64
+
+
 def futhark_hypot64(x, y):
     return np.hypot(x, y)
 
@@ -700,14 +736,12 @@ def futhark_isinf64(x):
     return np.isinf(x)
 
 
-def futhark_to_bits64(x):
-    s = struct.pack(">d", x)
-    return np.int64(struct.unpack(">q", s)[0])
+def fptobits_f64_i64(x):
+    return x.view(np.int64)
 
 
-def futhark_from_bits64(x):
-    s = struct.pack(">q", x)
-    return np.float64(struct.unpack(">d", s)[0])
+def bitstofp_i64_f64(x):
+    return x.view(np.float64)
 
 
 def futhark_log32(x):
@@ -730,6 +764,10 @@ def futhark_sqrt32(x):
     return np.float32(np.sqrt(x))
 
 
+def futhark_rsqrt32(x):
+    return np.float32(1 / np.sqrt(x))
+
+
 def futhark_cbrt32(x):
     return np.float32(np.cbrt(x))
 
@@ -742,26 +780,50 @@ def futhark_cos32(x):
     return np.cos(x)
 
 
+def futhark_cospi32(x):
+    return np.cos(pi32 * x)
+
+
 def futhark_sin32(x):
     return np.sin(x)
 
 
+def futhark_sinpi32(x):
+    return np.sin(pi32 * x)
+
+
 def futhark_tan32(x):
     return np.tan(x)
 
 
+def futhark_tanpi32(x):
+    return np.tan(pi32 * x)
+
+
 def futhark_acos32(x):
     return np.arccos(x)
 
 
+def futhark_acospi32(x):
+    return np.arccos(x) / pi32
+
+
 def futhark_asin32(x):
     return np.arcsin(x)
 
 
+def futhark_asinpi32(x):
+    return np.arcsin(x) / pi32
+
+
 def futhark_atan32(x):
     return np.arctan(x)
 
 
+def futhark_atanpi32(x):
+    return np.arctan(x) / pi32
+
+
 def futhark_cosh32(x):
     return np.cosh(x)
 
@@ -790,6 +852,10 @@ def futhark_atan2_32(x, y):
     return np.arctan2(x, y)
 
 
+def futhark_atan2pi_32(x, y):
+    return np.arctan2(x, y) / pi32
+
+
 def futhark_hypot32(x, y):
     return np.hypot(x, y)
 
@@ -834,14 +900,12 @@ def futhark_isinf32(x):
     return np.isinf(x)
 
 
-def futhark_to_bits32(x):
-    s = struct.pack(">f", x)
-    return np.int32(struct.unpack(">l", s)[0])
+def fptobits_f32_i32(x):
+    return x.view(np.int32)
 
 
-def futhark_from_bits32(x):
-    s = struct.pack(">l", x)
-    return np.float32(struct.unpack(">f", s)[0])
+def bitstofp_i32_f32(x):
+    return x.view(np.float32)
 
 
 def futhark_log16(x):
@@ -864,6 +928,10 @@ def futhark_sqrt16(x):
     return np.float16(np.sqrt(x))
 
 
+def futhark_rsqrt16(x):
+    return np.float16(1 / np.sqrt(x))
+
+
 def futhark_cbrt16(x):
     return np.float16(np.cbrt(x))
 
@@ -876,26 +944,50 @@ def futhark_cos16(x):
     return np.cos(x)
 
 
+def futhark_cospi16(x):
+    return np.cos(pi16 * x)
+
+
 def futhark_sin16(x):
     return np.sin(x)
 
 
+def futhark_sinpi16(x):
+    return np.sin(pi16 * x)
+
+
 def futhark_tan16(x):
     return np.tan(x)
 
 
+def futhark_tanpi16(x):
+    return np.tan(pi16 * x)
+
+
 def futhark_acos16(x):
     return np.arccos(x)
 
 
+def futhark_acospi16(x):
+    return np.arccos(x) / pi16
+
+
 def futhark_asin16(x):
     return np.arcsin(x)
 
 
+def futhark_asinpi16(x):
+    return np.arcsin(x) / pi16
+
+
 def futhark_atan16(x):
     return np.arctan(x)
 
 
+def futhark_atanpi16(x):
+    return np.arctan(x) / pi16
+
+
 def futhark_cosh16(x):
     return np.cosh(x)
 
@@ -924,6 +1016,10 @@ def futhark_atan2_16(x, y):
     return np.arctan2(x, y)
 
 
+def futhark_atan2pi_16(x, y):
+    return np.arctan2(x, y) / pi16
+
+
 def futhark_hypot16(x, y):
     return np.hypot(x, y)
 
@@ -968,14 +1064,12 @@ def futhark_isinf16(x):
     return np.isinf(x)
 
 
-def futhark_to_bits16(x):
-    s = struct.pack(">e", x)
-    return np.int16(struct.unpack(">H", s)[0])
+def fptobits_f16_i16(x):
+    return x.view(np.int16)
 
 
-def futhark_from_bits16(x):
-    s = struct.pack(">H", np.uint16(x))
-    return np.float16(struct.unpack(">e", s)[0])
+def bitstofp_i16_f16(x):
+    return x.view(np.float16)
 
 
 def futhark_lerp16(v0, v1, t):
