@@ -136,7 +136,7 @@ pdBinOp (FPow ft) a b =
   where
     derivs x y =
       ( y * (x ** (y - 1)),
-        condExp (x .<=. 0) 0 ((x ** y) * log x)
+        (x ** y) * condExp (x .<=. 0) 0 (log x)
       )
 pdBinOp (FMax ft) a b =
   floatBinOp derivs derivs derivs ft a b
@@ -159,6 +159,12 @@ pdBuiltin "sqrt32" [x] =
   Just [untyped $ 1 / (2 * sqrt (isF32 x))]
 pdBuiltin "sqrt64" [x] =
   Just [untyped $ 1 / (2 * sqrt (isF64 x))]
+pdBuiltin "rsqrt16" [x] =
+  Just [untyped $ -1 / (2 * (isF16 x ** (3 / 2)))]
+pdBuiltin "rsqrt32" [x] =
+  Just [untyped $ -1 / (2 * (isF32 x ** (3 / 2)))]
+pdBuiltin "rsqrt64" [x] =
+  Just [untyped $ -1 / (2 * (isF64 x ** (3 / 2)))]
 pdBuiltin "cbrt16" [x] =
   Just [untyped $ 1 / (3 * cbrt16 (isF16 x) * cbrt16 (isF16 x))]
   where
@@ -207,6 +213,12 @@ pdBuiltin "sin32" [x] =
   Just [untyped $ cos (isF32 x)]
 pdBuiltin "sin64" [x] =
   Just [untyped $ cos (isF64 x)]
+pdBuiltin "sinpi16" [x] =
+  Just [untyped $ pi * cos (pi * isF16 x)]
+pdBuiltin "sinpi32" [x] =
+  Just [untyped $ pi * cos (pi * isF32 x)]
+pdBuiltin "sinpi64" [x] =
+  Just [untyped $ pi * cos (pi * isF64 x)]
 pdBuiltin "sinh16" [x] =
   Just [untyped $ cosh (isF16 x)]
 pdBuiltin "sinh32" [x] =
@@ -219,6 +231,12 @@ pdBuiltin "cos32" [x] =
   Just [untyped $ -sin (isF32 x)]
 pdBuiltin "cos64" [x] =
   Just [untyped $ -sin (isF64 x)]
+pdBuiltin "cospi16" [x] =
+  Just [untyped $ -pi * sin (pi * isF16 x)]
+pdBuiltin "cospi32" [x] =
+  Just [untyped $ -pi * sin (pi * isF32 x)]
+pdBuiltin "cospi64" [x] =
+  Just [untyped $ -pi * sin (pi * isF64 x)]
 pdBuiltin "cosh16" [x] =
   Just [untyped $ sinh (isF16 x)]
 pdBuiltin "cosh32" [x] =
@@ -231,12 +249,24 @@ pdBuiltin "tan32" [x] =
   Just [untyped $ 1 / (cos (isF32 x) * cos (isF32 x))]
 pdBuiltin "tan64" [x] =
   Just [untyped $ 1 / (cos (isF64 x) * cos (isF64 x))]
+pdBuiltin "tanpi16" [x] =
+  Just [untyped $ pi * (1 / (cos (pi * isF16 x) * cos (pi * isF16 x)))]
+pdBuiltin "tanpi32" [x] =
+  Just [untyped $ pi * (1 / (cos (pi * isF32 x) * cos (pi * isF32 x)))]
+pdBuiltin "tanpi64" [x] =
+  Just [untyped $ pi * (1 / (cos (pi * isF64 x) * cos (pi * isF64 x)))]
 pdBuiltin "asin16" [x] =
   Just [untyped $ 1 / sqrt (1 - isF16 x * isF16 x)]
 pdBuiltin "asin32" [x] =
   Just [untyped $ 1 / sqrt (1 - isF32 x * isF32 x)]
 pdBuiltin "asin64" [x] =
   Just [untyped $ 1 / sqrt (1 - isF64 x * isF64 x)]
+pdBuiltin "asinpi16" [x] =
+  Just [untyped $ 1 / (pi * sqrt (1 - isF16 x * isF16 x))]
+pdBuiltin "asinpi32" [x] =
+  Just [untyped $ 1 / (pi * sqrt (1 - isF32 x * isF32 x))]
+pdBuiltin "asinpi64" [x] =
+  Just [untyped $ 1 / (pi * sqrt (1 - isF64 x * isF64 x))]
 pdBuiltin "asinh16" [x] =
   Just [untyped $ 1 / sqrt (1 + isF16 x * isF16 x)]
 pdBuiltin "asinh32" [x] =
@@ -249,6 +279,12 @@ pdBuiltin "acos32" [x] =
   Just [untyped $ -1 / sqrt (1 - isF32 x * isF32 x)]
 pdBuiltin "acos64" [x] =
   Just [untyped $ -1 / sqrt (1 - isF64 x * isF64 x)]
+pdBuiltin "acospi16" [x] =
+  Just [untyped $ -1 / (pi * sqrt (1 - isF16 x * isF16 x))]
+pdBuiltin "acospi32" [x] =
+  Just [untyped $ -1 / (pi * sqrt (1 - isF32 x * isF32 x))]
+pdBuiltin "acospi64" [x] =
+  Just [untyped $ -1 / (pi * sqrt (1 - isF64 x * isF64 x))]
 pdBuiltin "acosh16" [x] =
   Just [untyped $ 1 / sqrt (isF16 x * isF16 x - 1)]
 pdBuiltin "acosh32" [x] =
@@ -260,6 +296,12 @@ pdBuiltin "atan16" [x] =
 pdBuiltin "atan32" [x] =
   Just [untyped $ 1 / (1 + isF32 x * isF32 x)]
 pdBuiltin "atan64" [x] =
+  Just [untyped $ 1 / (pi * (1 + isF64 x * isF64 x))]
+pdBuiltin "atanpi16" [x] =
+  Just [untyped $ 1 / (pi * (1 + isF16 x * isF16 x))]
+pdBuiltin "atanpi32" [x] =
+  Just [untyped $ 1 / (pi * (1 + isF32 x * isF32 x))]
+pdBuiltin "atanpi64" [x] =
   Just [untyped $ 1 / (1 + isF64 x * isF64 x)]
 pdBuiltin "atanh16" [x] =
   Just [untyped $ cosh (isF16 x) * cosh (isF16 x)]
@@ -282,6 +324,21 @@ pdBuiltin "atan2_64" [x, y] =
     [ untyped $ -isF64 y / (isF64 x * isF64 x + isF64 y * isF64 y),
       untyped $ -isF64 x / (isF64 x * isF64 x + isF64 y * isF64 y)
     ]
+pdBuiltin "atan2pi_16" [x, y] =
+  Just
+    [ untyped $ -isF16 y / (pi * (isF16 x * isF16 x + isF16 y * isF16 y)),
+      untyped $ -isF16 x / (pi * (isF16 x * isF16 x + isF16 y * isF16 y))
+    ]
+pdBuiltin "atan2pi_32" [x, y] =
+  Just
+    [ untyped $ -isF32 y / (pi * (isF32 x * isF32 x + isF32 y * isF32 y)),
+      untyped $ -isF32 x / (pi * (isF32 x * isF32 x + isF32 y * isF32 y))
+    ]
+pdBuiltin "atan2pi_64" [x, y] =
+  Just
+    [ untyped $ -isF64 y / (pi * (isF64 x * isF64 x + isF64 y * isF64 y)),
+      untyped $ -isF64 x / (pi * (isF64 x * isF64 x + isF64 y * isF64 y))
+    ]
 pdBuiltin "tanh16" [x] =
   Just [untyped $ 1 - tanh (isF16 x) * tanh (isF16 x)]
 pdBuiltin "tanh32" [x] =
@@ -300,18 +357,6 @@ pdBuiltin "mad32" [a, b, _c] =
   Just [b, a, fConst Float32 1]
 pdBuiltin "mad64" [a, b, _c] =
   Just [b, a, fConst Float64 1]
-pdBuiltin "from_bits16" [_] =
-  Just [fConst Float16 1]
-pdBuiltin "from_bits32" [_] =
-  Just [fConst Float32 1]
-pdBuiltin "from_bits64" [_] =
-  Just [fConst Float64 1]
-pdBuiltin "to_bits16" [_] =
-  Just [iConst Int16 1]
-pdBuiltin "to_bits32" [_] =
-  Just [iConst Int32 1]
-pdBuiltin "to_bits64" [_] =
-  Just [iConst Int64 1]
 pdBuiltin "hypot16" [x, y] =
   Just
     [ untyped $ isF16 x / isF16 (FunExp "hypot16" [x, y] $ FloatType Float16),
