@@ -28,6 +28,15 @@ patName :: Pat Type -> ADM VName
 patName (Pat [pe]) = pure $ patElemName pe
 patName pat = error $ "Expected single-element pattern: " ++ prettyString pat
 
+copyIfArray :: VName -> ADM VName
+copyIfArray v = do
+  v_t <- lookupType v
+  case v_t of
+    Array {} ->
+      letExp (baseString v <> "_copy") . BasicOp $
+        Replicate mempty (Var v)
+    _ -> pure v
+
 -- The vast majority of BasicOps require no special treatment in the
 -- forward pass and produce one value (and hence one adjoint).  We
 -- deal with that case here.
@@ -42,28 +51,8 @@ commonBasicOp pat aux op m = do
 diffBasicOp :: Pat Type -> StmAux () -> BasicOp -> ADM () -> ADM ()
 diffBasicOp pat aux e m =
   case e of
-    CmpOp cmp x y -> do
-      (_pat_v, pat_adj) <- commonBasicOp pat aux e m
-      returnSweepCode $ do
-        let t = cmpOpType cmp
-            update contrib = do
-              void $ updateSubExpAdj x contrib
-              void $ updateSubExpAdj y contrib
-
-        case t of
-          FloatType ft ->
-            update <=< letExp "contrib" $
-              Match
-                [Var pat_adj]
-                [Case [Just $ BoolValue True] $ resultBody [constant (floatValue ft (1 :: Int))]]
-                (resultBody [constant (floatValue ft (0 :: Int))])
-                (MatchDec [Prim (FloatType ft)] MatchNormal)
-          IntType it ->
-            update <=< letExp "contrib" $ BasicOp $ ConvOp (BToI it) (Var pat_adj)
-          Bool ->
-            update pat_adj
-          Unit ->
-            pure ()
+    CmpOp {} ->
+      void $ commonBasicOp pat aux e m
     --
     ConvOp op x -> do
       (_pat_v, pat_adj) <- commonBasicOp pat aux e m
@@ -120,8 +109,7 @@ diffBasicOp pat aux e m =
     --
     Index arr slice -> do
       (_pat_v, pat_adj) <- commonBasicOp pat aux e m
-      returnSweepCode $ do
-        void $ updateAdjSlice slice arr pat_adj
+      returnSweepCode $ void $ updateAdjSlice slice arr pat_adj
     FlatIndex {} -> error "FlatIndex not handled by AD yet."
     FlatUpdate {} -> error "FlatUpdate not handled by AD yet."
     --
@@ -129,20 +117,20 @@ diffBasicOp pat aux e m =
       (_pat_v, pat_adj) <- commonBasicOp pat aux e m
       returnSweepCode $ updateSubExpAdj se pat_adj
     --
-    Reshape k _ arr -> do
+    Reshape arr newshape -> do
       (_pat_v, pat_adj) <- commonBasicOp pat aux e m
       returnSweepCode $ do
         arr_shape <- arrayShape <$> lookupType arr
         void $
           updateAdj arr <=< letExp "adj_reshape" . BasicOp $
-            Reshape k arr_shape pat_adj
+            Reshape pat_adj (reshapeAll (newShape newshape) arr_shape)
     --
-    Rearrange perm arr -> do
+    Rearrange arr perm -> do
       (_pat_v, pat_adj) <- commonBasicOp pat aux e m
       returnSweepCode $
         void $
           updateAdj arr <=< letExp "adj_rearrange" . BasicOp $
-            Rearrange (rearrangeInverse perm) pat_adj
+            Rearrange pat_adj (rearrangeInverse perm)
     --
     Replicate (Shape []) (Var se) -> do
       (_pat_v, pat_adj) <- commonBasicOp pat aux e m
@@ -157,7 +145,7 @@ diffBasicOp pat aux e m =
         n <- letSubExp "rep_size" =<< foldBinOp (Mul Int64 OverflowUndef) (intConst Int64 1) ns
         pat_adj_flat <-
           letExp (baseString pat_adj <> "_flat") . BasicOp $
-            Reshape ReshapeArbitrary (Shape $ n : arrayDims x_t) pat_adj
+            Reshape pat_adj (reshapeAll (Shape ns) (Shape $ n : arrayDims x_t))
         reduce <- reduceSOAC [Reduce Commutative lam [ne]]
         updateSubExpAdj x
           =<< letExp "rep_contrib" (Op $ Screma n [pat_adj_flat] reduce)
@@ -182,7 +170,7 @@ diffBasicOp pat aux e m =
 
         zipWithM_ updateAdj (arr : arrs) slices
     --
-    Manifest _ se -> do
+    Manifest se _ -> do
       (_pat_v, pat_adj) <- commonBasicOp pat aux e m
       returnSweepCode $ void $ updateAdj se pat_adj
     --
@@ -202,13 +190,7 @@ diffBasicOp pat aux e m =
       (_pat_v, pat_adj) <- commonBasicOp pat aux e m
       returnSweepCode $ do
         v_adj <- letExp "update_val_adj" $ BasicOp $ Index pat_adj slice
-        t <- lookupType v_adj
-        v_adj_copy <-
-          case t of
-            Array {} ->
-              letExp "update_val_adj_copy" . BasicOp $
-                Replicate mempty (Var v_adj)
-            _ -> pure v_adj
+        v_adj_copy <- copyIfArray v_adj
         updateSubExpAdj v v_adj_copy
         zeroes <- letSubExp "update_zero" . zeroExp =<< subExpType v
         void $
@@ -278,9 +260,18 @@ diffStm stm@(Let pat _ (Match ses cases defbody _)) m = do
           ses
           (map (fmap $ diffBody adjs branches_free) cases)
           (diffBody adjs branches_free defbody)
-    zipWithM_ insAdj branches_free branches_free_adj
+    -- See Note [Array Adjoints of Match]
+    forM_ (zip branches_free branches_free_adj) $ \(v, v_adj) ->
+      insAdj v =<< copyIfArray v_adj
 diffStm (Let pat aux (Op soac)) m =
-  vjpSOAC vjpOps pat aux soac m
+  -- We add the attributes from 'aux' to every SOAC (but only SOAC) produced. We
+  -- could do this on *every* stm, but it would be very verbose.
+  censorStms (fmap addAttrs) $ vjpSOAC vjpOps pat aux soac m
+  where
+    addAttrs stm
+      | Op _ <- stmExp stm =
+          attribute (stmAuxAttrs aux) stm
+      | otherwise = stm
 diffStm (Let pat aux loop@Loop {}) m =
   diffLoop diffStms pat aux loop m
 -- See Note [Adjoints of accumulators]
@@ -439,3 +430,43 @@ revVJP scope (Lambda params ts body) =
 -- our current translation rules, they will be dead code.  As long as
 -- we are careful to run dead code elimination after revVJP, we should
 -- be good.
+
+-- Note [Array Adjoints of Match]
+--
+-- Some unusual, but sadly not completely contrived, contain Match
+-- expressions that return multiple arrays, and there the arrays
+-- returned by one branch have overlapping aliases with another
+-- branch, although in different places. As an example consider this:
+--
+--   let (X,Y) = if c
+--               then (A, B)
+--               else (B, A)
+--
+-- Because our aliasing representation cannot express mutually
+-- exclusive aliases, we will consider X and Y to be aliased to each
+-- other. In practice, this means it is unlikely for X or Y to be
+-- consumed, because it would also consume the other (although it's
+-- possible for carefully written code).
+--
+-- When producing adjoints for this, it will be something like
+--
+--   let (X_adj,Y_adj) = if c
+--                       then (A_adj, B_adj)
+--                       else (B_adj, A_adj)
+--
+-- which completely reflects the primal code. However, while it is
+-- unlikely that any consumption takes place for the original primal
+-- variables, it is almost guaranteed that X_adj and Y_adj will be
+-- consumed (that is the main way we use adjoints after all), and due
+-- to the conservative aliasing, when one is consumed, so is the
+-- other! To avoid this tragic fate, we are forced to copy any
+-- array-typed adjoints returned by a Match. This can be quite costly.
+-- However:
+--
+-- 1) Futhark has pretty OK copy removal, so maybe it can get rid of
+--    these by using information not available to the AD pass.
+--
+-- 2) In many cases, arrays will have accumulator adjoints, which are
+--    not subject to this problem.
+--
+-- Issue #2228 was caused by neglecting to do this.
