@@ -100,28 +100,26 @@ import Futhark.Util.Pretty qualified as PP
 -- create a list, use 'ArrayTransforms' instead.
 data ArrayTransform
   = -- | A permutation of an otherwise valid input.
-    Rearrange Certs [Int]
+    Rearrange (StmAux ()) [Int]
   | -- | A reshaping of an otherwise valid input.
-    Reshape Certs ReshapeKind Shape
-  | -- | A reshaping of the outer dimension.
-    ReshapeOuter Certs ReshapeKind Shape
-  | -- | A reshaping of everything but the outer dimension.
-    ReshapeInner Certs ReshapeKind Shape
+    Reshape (StmAux ()) (NewShape SubExp)
   | -- | Replicate the rows of the array a number of times.
-    Replicate Certs Shape
+    Replicate (StmAux ()) Shape
   | -- | An array indexing operation.
-    Index Certs (Slice SubExp)
+    Index (StmAux ()) (Slice SubExp)
   deriving (Show, Eq, Ord)
 
+instance FreeIn ArrayTransform where
+  freeIn' (Rearrange cs _) = freeIn' cs
+  freeIn' (Reshape cs shape) = freeIn' cs <> freeIn' shape
+  freeIn' (Replicate cs shape) = freeIn' cs <> freeIn' shape
+  freeIn' (Index cs slice) = freeIn' cs <> freeIn' slice
+
 instance Substitute ArrayTransform where
   substituteNames substs (Rearrange cs xs) =
     Rearrange (substituteNames substs cs) xs
-  substituteNames substs (Reshape cs k ses) =
-    Reshape (substituteNames substs cs) k (substituteNames substs ses)
-  substituteNames substs (ReshapeOuter cs k ses) =
-    ReshapeOuter (substituteNames substs cs) k (substituteNames substs ses)
-  substituteNames substs (ReshapeInner cs k ses) =
-    ReshapeInner (substituteNames substs cs) k (substituteNames substs ses)
+  substituteNames substs (Reshape cs newshape) =
+    Reshape (substituteNames substs cs) (substituteNames substs newshape)
   substituteNames substs (Replicate cs se) =
     Replicate (substituteNames substs cs) (substituteNames substs se)
   substituteNames substs (Index cs slice) =
@@ -153,6 +151,9 @@ instance Substitute ArrayTransforms where
   substituteNames substs (ArrayTransforms ts) =
     ArrayTransforms $ substituteNames substs <$> ts
 
+instance FreeIn ArrayTransforms where
+  freeIn' (ArrayTransforms trs) = foldMap freeIn' trs
+
 -- | The empty transformation list.
 noTransforms :: ArrayTransforms
 noTransforms = ArrayTransforms Seq.empty
@@ -231,11 +232,11 @@ combineTransforms _ _ = Nothing
 -- an input transformation of an array variable.  If so, return the
 -- variable and the transformation.  Only 'Rearrange' and 'Reshape'
 -- are possible to express this way.
-transformFromExp :: Certs -> Exp rep -> Maybe (VName, ArrayTransform)
-transformFromExp cs (BasicOp (Futhark.Rearrange perm v)) =
+transformFromExp :: StmAux () -> Exp rep -> Maybe (VName, ArrayTransform)
+transformFromExp cs (BasicOp (Futhark.Rearrange v perm)) =
   Just (v, Rearrange cs perm)
-transformFromExp cs (BasicOp (Futhark.Reshape k shape v)) =
-  Just (v, Reshape cs k shape)
+transformFromExp cs (BasicOp (Futhark.Reshape v shape)) =
+  Just (v, Reshape cs shape)
 transformFromExp cs (BasicOp (Futhark.Replicate shape (Var v))) =
   Just (v, Replicate cs shape)
 transformFromExp cs (BasicOp (Futhark.Index v slice)) =
@@ -243,20 +244,14 @@ transformFromExp cs (BasicOp (Futhark.Index v slice)) =
 transformFromExp _ _ = Nothing
 
 -- | Turn an array transform on an array back into an expression.
-transformToExp :: (Monad m, HasScope rep m) => ArrayTransform -> VName -> m (Certs, Exp rep)
+transformToExp :: (Monad m, HasScope rep m) => ArrayTransform -> VName -> m (StmAux (), Exp rep)
 transformToExp (Replicate cs n) ia =
   pure (cs, BasicOp $ Futhark.Replicate n (Var ia))
 transformToExp (Rearrange cs perm) ia = do
   r <- arrayRank <$> lookupType ia
-  pure (cs, BasicOp $ Futhark.Rearrange (perm ++ [length perm .. r - 1]) ia)
-transformToExp (Reshape cs k shape) ia = do
-  pure (cs, BasicOp $ Futhark.Reshape k shape ia)
-transformToExp (ReshapeOuter cs k shape) ia = do
-  shape' <- reshapeOuter shape 1 . arrayShape <$> lookupType ia
-  pure (cs, BasicOp $ Futhark.Reshape k shape' ia)
-transformToExp (ReshapeInner cs k shape) ia = do
-  shape' <- reshapeInner shape 1 . arrayShape <$> lookupType ia
-  pure (cs, BasicOp $ Futhark.Reshape k shape' ia)
+  pure (cs, BasicOp $ Futhark.Rearrange ia (perm ++ [length perm .. r - 1]))
+transformToExp (Reshape cs shape) ia = do
+  pure (cs, BasicOp $ Futhark.Reshape ia shape)
 transformToExp (Index cs slice) ia = do
   pure (cs, BasicOp $ Futhark.Index ia slice)
 
@@ -268,6 +263,9 @@ transformToExp (Index cs slice) ia = do
 data Input = Input ArrayTransforms VName Type
   deriving (Show, Eq, Ord)
 
+instance FreeIn Input where
+  freeIn' (Input trs v t) = freeIn' trs <> freeIn' v <> freeIn' t
+
 instance Substitute Input where
   substituteNames substs (Input ts v t) =
     Input
@@ -296,8 +294,10 @@ isVarInput _ = Nothing
 isVarishInput :: Input -> Maybe VName
 isVarishInput (Input ts v t)
   | nullTransforms ts = Just v
-  | Reshape cs ReshapeCoerce (Shape [_]) :< ts' <- viewf ts,
-    cs == mempty =
+  | Reshape aux newshape :< ts' <- viewf ts,
+    ReshapeCoerce <- reshapeKind newshape,
+    1 <- shapeRank $ newShape newshape,
+    stmAuxCerts aux == mempty =
       isVarishInput $ Input ts' v t
 isVarishInput _ = Nothing
 
@@ -313,15 +313,13 @@ addInitialTransforms ts (Input ots a t) = Input (ts <> ots) a t
 
 applyTransform :: (MonadBuilder m) => ArrayTransform -> VName -> m VName
 applyTransform tr ia = do
-  (cs, e) <- transformToExp tr ia
-  certifying cs $ letExp s e
+  (aux, e) <- transformToExp tr ia
+  auxing aux $ letExp s e
   where
     s = case tr of
       Replicate {} -> "replicate"
       Rearrange {} -> "rearrange"
       Reshape {} -> "reshape"
-      ReshapeOuter {} -> "reshape_outer"
-      ReshapeInner {} -> "reshape_inner"
       Index {} -> "index"
 
 applyTransforms :: (MonadBuilder m) => ArrayTransforms -> VName -> m VName
@@ -353,14 +351,8 @@ inputType (Input (ArrayTransforms ts) _ at) =
       arrayOfShape t shape
     transformType t (Rearrange _ perm) =
       rearrangeType perm t
-    transformType t (Reshape _ _ shape) =
-      t `setArrayShape` shape
-    transformType t (ReshapeOuter _ _ shape) =
-      let Shape oldshape = arrayShape t
-       in t `setArrayShape` Shape (shapeDims shape ++ drop 1 oldshape)
-    transformType t (ReshapeInner _ _ shape) =
-      let Shape oldshape = arrayShape t
-       in t `setArrayShape` Shape (take 1 oldshape ++ shapeDims shape)
+    transformType t (Reshape _ shape) =
+      t `setArrayShape` newShape shape
     transformType t (Index _ slice) =
       t `setArrayShape` sliceShape slice
 
@@ -380,8 +372,11 @@ transformRows (ArrayTransforms ts) =
   where
     transformRows' inp (Rearrange cs perm) =
       addTransform (Rearrange cs (0 : map (+ 1) perm)) inp
-    transformRows' inp (Reshape cs k shape) =
-      addTransform (ReshapeInner cs k shape) inp
+    transformRows' inp (Reshape cs shape) =
+      addTransform (Reshape cs newshape) inp
+      where
+        newshape = reshapeAll inp_shape $ Shape [shapeSize 0 inp_shape] <> newShape shape
+        inp_shape = arrayShape $ inputType inp
     transformRows' inp (Replicate cs n)
       | inputRank inp == 1 =
           Rearrange mempty [1, 0]
@@ -700,18 +695,8 @@ soacToStream soac = do
 ppArrayTransform :: PP.Doc a -> ArrayTransform -> PP.Doc a
 ppArrayTransform e (Rearrange cs perm) =
   "rearrange" <> pretty cs <> PP.apply [PP.apply (map pretty perm), e]
-ppArrayTransform e (Reshape cs ReshapeArbitrary shape) =
+ppArrayTransform e (Reshape cs shape) =
   "reshape" <> pretty cs <> PP.apply [pretty shape, e]
-ppArrayTransform e (ReshapeOuter cs ReshapeArbitrary shape) =
-  "reshape_outer" <> pretty cs <> PP.apply [pretty shape, e]
-ppArrayTransform e (ReshapeInner cs ReshapeArbitrary shape) =
-  "reshape_inner" <> pretty cs <> PP.apply [pretty shape, e]
-ppArrayTransform e (Reshape cs ReshapeCoerce shape) =
-  "coerce" <> pretty cs <> PP.apply [pretty shape, e]
-ppArrayTransform e (ReshapeOuter cs ReshapeCoerce shape) =
-  "coerce_outer" <> pretty cs <> PP.apply [pretty shape, e]
-ppArrayTransform e (ReshapeInner cs ReshapeCoerce shape) =
-  "coerce_inner" <> pretty cs <> PP.apply [pretty shape, e]
 ppArrayTransform e (Replicate cs ne) =
   "replicate" <> pretty cs <> PP.apply [pretty ne, e]
 ppArrayTransform e (Index cs slice) =
