@@ -36,6 +36,7 @@ module Futhark.CodeGen.Backends.GenericC.Monad
     items,
     stm,
     stms,
+    comment,
     decl,
     headerDecl,
     publicDef,
@@ -60,6 +61,9 @@ module Futhark.CodeGen.Backends.GenericC.Monad
     collect',
     contextType,
     configType,
+    localProvenance,
+    askProvenance,
+    provenanceExp,
 
     -- * Building Blocks
     copyMemoryDefaultSpace,
@@ -258,7 +262,9 @@ data CompilerEnv op s = CompilerEnv
     -- pressure, we keep these allocations around for a long time, and
     -- record their sizes so we can reuse them if possible (and
     -- realloc() when needed).
-    envCachedMem :: M.Map C.Exp VName
+    envCachedMem :: M.Map C.Exp VName,
+    -- | The provenance of an enclosing 'Imp.MetaProvenance', if any.
+    envProvenance :: Provenance
   }
 
 contextContents :: CompilerM op s ([C.FieldGroup], [C.Stm], [C.Stm])
@@ -322,7 +328,7 @@ runCompilerM ::
   (a, CompilerState s)
 runCompilerM ops src userstate (CompilerM m) =
   runState
-    (runReaderT m (CompilerEnv ops mempty))
+    (runReaderT m (CompilerEnv ops mempty mempty))
     (newCompilerState src userstate)
 
 getUserState :: CompilerM op s s
@@ -344,6 +350,23 @@ collect' m = do
   modify $ \s -> s {compItems = old}
   pure (x, DL.toList new)
 
+-- | Locally replace (not extend!) the provenance.
+localProvenance :: Provenance -> CompilerM op s a -> CompilerM op s a
+localProvenance p = local $ \env -> env {envProvenance = p}
+
+-- | The provenance of the closest enclosing 'Imp.MetaProvenance'.
+askProvenance :: CompilerM op s Provenance
+askProvenance = asks envProvenance
+
+-- | A C expression corresponding to the current provenance.
+provenanceExp :: CompilerM op s C.Exp
+provenanceExp = do
+  p <- askProvenance
+  pure $
+    if p == mempty
+      then [C.cexp|NULL|]
+      else [C.cexp|$string:(prettyString p)|]
+
 -- | Used when we, inside an existing 'CompilerM' action, want to
 -- generate code for a new function.  Use this so that the compiler
 -- understands that previously declared memory doesn't need to be
@@ -358,12 +381,20 @@ inNewFunction m = do
   where
     noCached env = env {envCachedMem = mempty}
 
+-- | Insert a block item in the generated code at this point.
 item :: C.BlockItem -> CompilerM op s ()
 item x = modify $ \s -> s {compItems = DL.snoc (compItems s) x}
 
 items :: [C.BlockItem] -> CompilerM op s ()
 items xs = modify $ \s -> s {compItems = DL.append (compItems s) (DL.fromList xs)}
 
+-- | Insert a comment in the generated code at this point. The comment may
+-- contain linebreaks, and must not contain any comment markers.
+comment :: T.Text -> CompilerM op s ()
+comment = mapM_ (f . ("// " <>)) . T.lines
+  where
+    f s = stm [C.cstm|$escstm:(T.unpack s)|]
+
 fatMemory :: Space -> CompilerM op s Bool
 fatMemory ScalarSpace {} = pure False
 fatMemory _ = asks $ opsFatMemory . envOperations
@@ -425,6 +456,7 @@ onClear :: C.BlockItem -> CompilerM op s ()
 onClear x = modify $ \s ->
   s {compClearItems = compClearItems s <> DL.singleton x}
 
+-- | Insert a statement in the generated code at this point.
 stm :: C.Stm -> CompilerM op s ()
 stm s = item [C.citem|$stm:s|]
 
