@@ -474,13 +474,8 @@ compileExp (FunExp h args _) = do
 -- All recursive constructors are duplicated here, since not doing so
 -- would cause use to enter regular generic C codegen with no escape.
 compileCode :: MCCode -> ISPCCompilerM ()
-compileCode (Comment s code) = do
-  xs <- GC.collect $ compileCode code
-  let comment = "// " ++ T.unpack s
-  GC.stm
-    [C.cstm|$comment:comment
-              { $items:xs }
-             |]
+compileCode (Meta (MetaComment s)) = do
+  GC.comment s
 compileCode (DeclareScalar name _ t) = do
   let ct = GC.primTypeToCType t
   quals <- getVariabilityQuals name
@@ -729,7 +724,6 @@ mayProduceError (x :>>: y) = mayProduceError x || mayProduceError y
 mayProduceError (If _ x y) = mayProduceError x || mayProduceError y
 mayProduceError (For _ _ x) = mayProduceError x
 mayProduceError (While _ x) = mayProduceError x
-mayProduceError (Comment _ x) = mayProduceError x
 mayProduceError (Op (ForEachActive _ body)) = mayProduceError body
 mayProduceError (Op (ForEach _ _ _ body)) = mayProduceError body
 mayProduceError (Op SegOp {}) = True
@@ -1015,8 +1009,6 @@ findDeps (For idx bound x) = do
     free = freeIn bound
 findDeps (While cond x) = do
   local (<> freeIn cond) $ findDeps x
-findDeps (Comment _ x) =
-  findDeps x
 findDeps (Op (SegOp _ free _ _ retvals _)) =
   mapM_
     ( \x ->
@@ -1064,7 +1056,6 @@ findVarying (x :>>: y) = findVarying x ++ findVarying y
 findVarying (If _ x y) = findVarying x ++ findVarying y
 findVarying (For _ _ x) = findVarying x
 findVarying (While _ x) = findVarying x
-findVarying (Comment _ x) = findVarying x
 findVarying (Op (ForEachActive _ body)) = findVarying body
 findVarying (Op (ForEach idx _ _ body)) = idx : findVarying body
 findVarying (DeclareMem mem _) = [mem]
