@@ -365,7 +365,7 @@ compileBlockOp pat (Inner (SegOp (SegMap lvl space _ body))) = do
       zipWithM_ (compileThreadResult space) (patElems pat) $
         kernelBodyResult body
   sOp $ Imp.ErrorSync Imp.FenceLocal
-compileBlockOp pat (Inner (SegOp (SegScan lvl space scans _ body))) = do
+compileBlockOp pat (Inner (SegOp (SegScan lvl space _ body scans))) = do
   compileFlatId space
 
   let (ltids, dims) = unzip $ unSegSpace space
@@ -412,7 +412,7 @@ compileBlockOp pat (Inner (SegOp (SegScan lvl space scans _ body))) = do
         (product dims')
         (segBinOpLambda scan)
         arrs_flat
-compileBlockOp pat (Inner (SegOp (SegRed lvl space ops _ body))) = do
+compileBlockOp pat (Inner (SegOp (SegRed lvl space _ body ops))) = do
   compileFlatId space
 
   let dims' = map pe64 dims
@@ -533,7 +533,7 @@ compileBlockOp pat (Inner (SegOp (SegRed lvl space ops _ body))) = do
             (map (unitSlice 0) (init dims') ++ [DimFix $ last dims' - 1])
 
       sOp $ Imp.Barrier Imp.FenceLocal
-compileBlockOp pat (Inner (SegOp (SegHist lvl space ops _ kbody))) = do
+compileBlockOp pat (Inner (SegOp (SegHist lvl space _ kbody ops))) = do
   compileFlatId space
   let (ltids, dims) = unzip $ unSegSpace space
 
@@ -578,6 +578,7 @@ compileBlockOp pat (Inner (SegOp (SegHist lvl space ops _ kbody))) = do
 compileBlockOp pat _ =
   compilerBugS $ "compileBlockOp: cannot compile rhs of binding " ++ prettyString pat
 
+-- | The operations for block-level kernels.
 blockOperations :: Operations GPUMem KernelEnv Imp.KernelOp
 blockOperations =
   (defaultOperations compileBlockOp)
@@ -598,6 +599,7 @@ arrayInSharedMemory (Var name) = do
     _ -> pure False
 arrayInSharedMemory Constant {} = pure False
 
+-- | Create a kernel with GPU operations at the block level.
 sKernelBlock ::
   String ->
   VName ->
@@ -606,6 +608,8 @@ sKernelBlock ::
   CallKernelGen ()
 sKernelBlock = sKernel blockOperations $ sExt64 . kernelBlockId
 
+-- | Generate code for writing this result of a block-level kernel. The
+-- 'PatElem' and 'KernelResult' must be matched as in the original segop.
 compileBlockResult ::
   SegSpace ->
   PatElem LetDecMem ->
