@@ -27,6 +27,7 @@ import Data.Char (isAlpha)
 import Data.Functor
 import Data.List (singleton)
 import Data.List.NonEmpty (NonEmpty (..))
+import Data.Loc qualified as L
 import Data.Maybe
 import Data.Set qualified as S
 import Data.Text qualified as T
@@ -238,11 +239,30 @@ pErrorMsgPart =
 pErrorMsg :: Parser (ErrorMsg SubExp)
 pErrorMsg = ErrorMsg <$> braces (pErrorMsgPart `sepBy` pComma)
 
-pSrcLoc :: Parser SrcLoc
-pSrcLoc = pStringLiteral $> mempty -- FIXME
+lInt :: Parser Int
+lInt = lexeme L.decimal
 
-pErrorLoc :: Parser (SrcLoc, [SrcLoc])
-pErrorLoc = (,mempty) <$> pSrcLoc
+pPos :: Parser L.Pos
+pPos =
+  L.Pos
+    <$> manyTill L.charLiteral (char ':')
+    <*> lInt
+    <*> (char ':' *> lInt)
+    <*> pure (-1) -- Cannot recover offset.
+
+pLocStr :: Parser Loc
+pLocStr = do
+  start <- pPos
+  void $ char '-'
+  end <- do
+    x <- lInt
+    choice
+      [ do
+          y <- char ':' *> lInt
+          pure $ L.Pos (L.posFile start) x y (-1),
+        pure $ L.Pos (L.posFile start) (L.posLine start) x (-1)
+      ]
+  pure $ L.Loc start end
 
 pIota :: Parser BasicOp
 pIota =
@@ -260,6 +280,12 @@ pIota =
               <*> pure t
           )
 
+pDimSplice :: Parser (DimSplice SubExp)
+pDimSplice = DimSplice <$> pInt <* lexeme "::" <*> pInt <* lexeme "=>" <*> pShape
+
+pNewShape :: Parser (NewShape SubExp)
+pNewShape = parens $ NewShape <$> (pDimSplice `sepBy` pComma) <* pSemi <*> pShape
+
 pBasicOp :: Parser BasicOp
 pBasicOp =
   choice
@@ -268,29 +294,19 @@ pBasicOp =
         $> uncurry (Opaque . OpaqueTrace)
         <*> parens ((,) <$> pStringLiteral <* pComma <*> pSubExp),
       keyword "copy" $> Replicate mempty . Var <*> parens pVName,
-      keyword "assert"
-        *> parens
-          ( Assert
-              <$> pSubExp
-              <* pComma
-              <*> pErrorMsg
-              <* pComma
-              <*> pErrorLoc
-          ),
+      keyword "assert" *> parens (Assert <$> pSubExp <* pComma <*> pErrorMsg),
       keyword "replicate"
         *> parens (Replicate <$> pShape <* pComma <*> pSubExp),
       keyword "reshape"
-        *> parens (Reshape ReshapeArbitrary <$> pShape <* pComma <*> pVName),
-      keyword "coerce"
-        *> parens (Reshape ReshapeCoerce <$> pShape <* pComma <*> pVName),
+        *> parens (Reshape <$> pVName <* pComma <*> pNewShape),
       keyword "scratch"
         *> parens (Scratch <$> pPrimType <*> many (pComma *> pSubExp)),
       keyword "rearrange"
         *> parens
-          (Rearrange <$> parens (pInt `sepBy` pComma) <* pComma <*> pVName),
+          (Rearrange <$> pVName <* pComma <*> parens (pInt `sepBy` pComma)),
       keyword "manifest"
         *> parens
-          (Manifest <$> parens (pInt `sepBy` pComma) <* pComma <*> pVName),
+          (Manifest <$> pVName <* pComma <*> parens (pInt `sepBy` pComma)),
       keyword "concat" *> do
         d <- "@" *> L.decimal
         parens $ do
@@ -337,6 +353,8 @@ pBasicOp =
       pConvOp "btoi" (const BToI) (keyword "bool") pIntType,
       pConvOp "ftob" (const . FToB) pFloatType (keyword "bool"),
       pConvOp "btof" (const BToF) (keyword "bool") pFloatType,
+      pConvOp "fptobits" (const . FPToBits) pFloatType pIntType,
+      pConvOp "bitstofp" (const BitsToFP) pIntType pFloatType,
       --
       pIndex,
       pFlatIndex,
@@ -481,7 +499,7 @@ pApply pr =
         <*> parens (pArg `sepBy` pComma)
         <* pColon
         <*> pRetTypes pr
-        <*> pure (safety, mempty, mempty)
+        <*> pure safety
 
     pArg =
       choice
@@ -583,11 +601,21 @@ pCerts =
 pSubExpRes :: Parser SubExpRes
 pSubExpRes = SubExpRes <$> pCerts <*> pSubExp
 
+pProvenance :: Parser Provenance
+pProvenance =
+  choice
+    [ lexeme $ between (char '"') (char '"') $ do
+        l <- pLocStr `sepBy1` "->"
+        pure $ Provenance (init l) (last l),
+      pure mempty
+    ]
+
 pStm :: PR rep -> Parser (Stm rep)
-pStm pr =
-  keyword "let" $> Let <*> pPat pr <* pEqual <*> pStmAux <*> pExp pr
+pStm pr = do
+  loc <- pProvenance
+  keyword "let" $> Let <*> pPat pr <* pEqual <*> pStmAux loc <*> pExp pr
   where
-    pStmAux = flip StmAux <$> pAttrs <*> pCerts <*> pure (pExpDec pr)
+    pStmAux loc = flip StmAux <$> pAttrs <*> pCerts <*> pure loc <*> pure (pExpDec pr)
 
 pStms :: PR rep -> Parser (Stms rep)
 pStms pr = stmsFromList <$> many (pStm pr)
@@ -919,18 +947,10 @@ pSegOp pr pLvl =
       keyword "seghist" *> pSegHist
     ]
   where
-    pSegMap =
-      SegOp.SegMap
-        <$> pLvl
-        <*> pSegSpace
-        <* pColon
-        <*> pTypes
-        <*> braces (pKernelBody pr)
-    pSegOp' f p =
+    pSegOp' f =
       f
         <$> pLvl
         <*> pSegSpace
-        <*> parens (p `sepBy` pComma)
         <* pColon
         <*> pTypes
         <*> braces (pKernelBody pr)
@@ -953,9 +973,10 @@ pSegOp pr pLvl =
         <*> pShape
         <* pComma
         <*> pLambda pr
-    pSegRed = pSegOp' SegOp.SegRed pSegBinOp
-    pSegScan = pSegOp' SegOp.SegScan pSegBinOp
-    pSegHist = pSegOp' SegOp.SegHist pHistOp
+    pSegMap = pSegOp' SegOp.SegMap
+    pSegRed = pSegOp' SegOp.SegRed <*> parens (pSegBinOp `sepBy` pComma)
+    pSegScan = pSegOp' SegOp.SegScan <*> parens (pSegBinOp `sepBy` pComma)
+    pSegHist = pSegOp' SegOp.SegHist <*> parens (pHistOp `sepBy` pComma)
 
 pSegLevel :: Parser GPU.SegLevel
 pSegLevel =
