@@ -1,9 +1,8 @@
 {-# OPTIONS_GHC -fno-warn-orphans #-}
 
--- | Futhark prettyprinter.  This module defines 'Pretty' instances
--- for the AST defined in "Futhark.IR.Syntax",
--- but also a number of convenience functions if you don't want to use
--- the interface from 'Pretty'.
+-- | Futhark prettyprinter. This module defines 'Pretty' instances for the AST
+-- defined in "Futhark.IR.Syntax", but also a number of convenience functions if
+-- you don't want to use the interface from 'Pretty'.
 module Futhark.IR.Pretty
   ( prettyTuple,
     prettyTupleLines,
@@ -43,7 +42,7 @@ instance Pretty Commutativity where
   pretty Commutative = "commutative"
   pretty Noncommutative = "noncommutative"
 
-instance Pretty Shape where
+instance (Pretty d) => Pretty (ShapeBase d) where
   pretty = mconcat . map (brackets . pretty) . shapeDims
 
 instance Pretty Rank where
@@ -53,9 +52,6 @@ instance (Pretty a) => Pretty (Ext a) where
   pretty (Free e) = pretty e
   pretty (Ext x) = "?" <> pretty (show x)
 
-instance Pretty ExtShape where
-  pretty = mconcat . map (brackets . pretty) . shapeDims
-
 instance Pretty Space where
   pretty DefaultSpace = mempty
   pretty (Space s) = "@" <> pretty s
@@ -152,6 +148,16 @@ stmCertAnnots = certAnnots . stmAuxCerts . stmAux
 instance Pretty Attrs where
   pretty = hsep . attrAnnots
 
+prettyLoc :: Loc -> Doc a
+prettyLoc = pretty . locText
+
+instance Pretty Provenance where
+  pretty (Provenance locs loc) = mconcat $ punctuate "->" $ map prettyLoc $ locs ++ [loc]
+
+instance (Pretty dec) => Pretty (StmAux dec) where
+  pretty (StmAux cs attrs p dec) =
+    braces $ mconcat $ punctuate semi [pretty cs, pretty attrs, pretty p, pretty dec]
+
 instance (Pretty t) => Pretty (Pat t) where
   pretty (Pat xs) = braces $ commastack $ map pretty xs
 
@@ -164,13 +170,18 @@ instance (Pretty t) => Pretty (Param t) where
 
 instance (PrettyRep rep) => Pretty (Stm rep) where
   pretty stm@(Let pat aux e) =
-    align . hang 2 $
+    (locstr <>) . align . hang 2 $
       "let"
         <+> align (pretty pat)
         <+> case stmannot of
           [] -> equals </> pretty e
           _ -> equals </> (stack stmannot </> pretty e)
     where
+      locstr =
+        if stmAuxLoc aux == mempty
+          then mempty
+          else dquotes (pretty (stmAuxLoc aux)) <> line
+
       stmannot =
         concat
           [ maybeToList (ppExpDec (stmAuxDec aux) e),
@@ -187,6 +198,13 @@ instance (Pretty d) => Pretty (FlatDimIndex d) where
 instance (Pretty a) => Pretty (FlatSlice a) where
   pretty (FlatSlice offset xs) = brackets (pretty offset <> ";" <+> commasep (map pretty xs))
 
+instance (Pretty d) => Pretty (DimSplice d) where
+  pretty (DimSplice i k shape) = pretty i <> "::" <> pretty k <> "=>" <> pretty shape
+
+instance (Pretty d) => Pretty (NewShape d) where
+  pretty (NewShape ds shape) =
+    parens $ align $ commastack (map pretty ds) <> semi </> pretty shape
+
 instance Pretty BasicOp where
   pretty (SubExp se) = pretty se
   pretty (Opaque OpaqueNil e) = "opaque" <> apply [pretty e]
@@ -228,17 +246,16 @@ instance Pretty BasicOp where
     "replicate" <> apply [pretty ne, align (pretty ve)]
   pretty (Scratch t shape) =
     "scratch" <> apply (pretty t : map pretty shape)
-  pretty (Reshape ReshapeArbitrary shape e) =
-    "reshape" <> apply [pretty shape, pretty e]
-  pretty (Reshape ReshapeCoerce shape e) =
-    "coerce" <> apply [pretty shape, pretty e]
-  pretty (Rearrange perm e) =
-    "rearrange" <> apply [apply (map pretty perm), pretty e]
+  pretty (Reshape reshape e) =
+    "reshape" <> parens (align $ commastack [pretty reshape, pretty e])
+  pretty (Rearrange v perm) =
+    "rearrange" <> apply [pretty v, apply (map pretty perm)]
   pretty (Concat i (x :| xs) w) =
     "concat" <> "@" <> pretty i <> apply (pretty w : pretty x : map pretty xs)
-  pretty (Manifest perm e) = "manifest" <> apply [apply (map pretty perm), pretty e]
-  pretty (Assert e msg (loc, _)) =
-    "assert" <> apply [pretty e, pretty msg, pretty $ show $ locStr loc]
+  pretty (Manifest v perm) =
+    "manifest" <> apply [pretty v, apply (map pretty perm)]
+  pretty (Assert e msg) =
+    "assert" <> apply [pretty e, pretty msg]
   pretty (UpdateAcc safety acc is v) =
     update_acc_str
       <> apply
@@ -306,12 +323,12 @@ instance (PrettyRep rep) => Pretty (Exp rep) where
         MatchFallback -> " <fallback>"
         MatchEquiv -> " <equiv>"
   pretty (BasicOp op) = pretty op
-  pretty (Apply fname args ret (safety, _, _)) =
+  pretty (Apply fname args ret safety) =
     applykw
       <+> pretty (nameToString fname)
       <> apply (map (align . prettyArg) args)
         </> colon
-        <+> braces (commasep $ map prettyRet ret)
+        <+> braces (align $ commasep $ map prettyRet ret)
     where
       prettyArg (arg, Consume) = "*" <> pretty arg
       prettyArg (arg, _) = pretty arg
