@@ -101,23 +101,19 @@ basicOpType (Replicate shape e) =
   pure . flip arrayOfShape shape <$> subExpType e
 basicOpType (Scratch t shape) =
   pure [arrayOf (Prim t) (Shape shape) NoUniqueness]
-basicOpType (Reshape _ (Shape []) e) =
+basicOpType (Reshape e shape) =
   result <$> lookupType e
   where
-    result t = [Prim $ elemType t]
-basicOpType (Reshape _ shape e) =
-  result <$> lookupType e
-  where
-    result t = [t `setArrayShape` shape]
-basicOpType (Rearrange perm e) =
-  result <$> lookupType e
+    result t = [t `setArrayShape` newShape shape]
+basicOpType (Rearrange v perm) =
+  result <$> lookupType v
   where
     result t = [rearrangeType perm t]
 basicOpType (Concat i (x :| _) ressize) =
   result <$> lookupType x
   where
     result xt = [setDimSize i xt ressize]
-basicOpType (Manifest _ v) =
+basicOpType (Manifest v _) =
   pure <$> lookupType v
 basicOpType Assert {} =
   pure [Prim Unit]
