@@ -92,8 +92,8 @@ finalizeNode nt = case nt of
       letBindNames [output] . BasicOp . SubExp . Var =<< H.applyTransforms ots v
   ResNode _ -> pure mempty
   TransNode output tr ia -> do
-    (cs, e) <- H.transformToExp tr ia
-    runBuilder_ $ certifying cs $ letBindNames [output] e
+    (aux, e) <- H.transformToExp tr ia
+    runBuilder_ $ auxing aux $ letBindNames [output] e
   FreeNode _ -> pure mempty
   DoNode stm lst -> do
     lst' <- mapM (finalizeNode . fst) lst
@@ -350,7 +350,6 @@ vFuseNodeT
       pat1_acc_nms <- namesFromList $ take n $ map patElemName $ patElems pat1,
       -- not $ namesIntersect (freeIn soac) pat1_acc_nms
       all ((`notNameIn` pat1_acc_nms) . getName) edges = do
-        let empty_aux = StmAux mempty mempty mempty
         wlam <- fst <$> doFusionInLambda wlam0
         bdy' <-
           runBodyBuilder $ inScopeOf wlam $ do
@@ -360,7 +359,7 @@ vFuseNodeT
             let other_pr1 = drop n $ zip (patElems pat1) wlam_res
             forM_ other_pr1 $ \(pat_elm, bdy_res) -> do
               let (nm, se, tp) = (patElemName pat_elm, resSubExp bdy_res, patElemType pat_elm)
-                  aux = empty_aux {stmAuxCerts = resCerts bdy_res}
+                  aux = (defAux ()) {stmAuxCerts = resCerts bdy_res}
               addStm $ Let (Pat [PatElem nm tp]) aux $ BasicOp $ SubExp se
             -- add the soac stmt
             soac' <- H.toExp soac
@@ -387,18 +386,19 @@ vFuseNodeT
     | wacc2_cons_nms <- namesFromList $ concatMap (\(_, nms, _) -> nms) w2_inps,
       wacc1_indep_nms <- map getName is1,
       all (`notNameIn` wacc2_cons_nms) wacc1_indep_nms = do
-        -- \^ the other safety checks are done inside `tryFuseWithAccs`
+        -- the other safety checks are done inside `tryFuseWithAccs`
         lam1' <- fst <$> doFusionInLambda lam1
         lam2' <- fst <$> doFusionInLambda lam2
         let stm1 = Let pat1 aux1 (WithAcc w1_inps lam1')
             stm2 = Let pat2 aux2 (WithAcc w2_inps lam2')
-        mstm <- SF.tryFuseWithAccs infusible stm1 stm2
+        mstm <- sequence $ SF.tryFuseWithAccs infusible stm1 stm2
         case mstm of
+          Nothing -> pure Nothing
           Just (Let pat aux (WithAcc w_inps wlam)) -> do
             (wlam', success) <- doFusionInLambda wlam
             let new_stm = Let pat aux (WithAcc w_inps wlam')
             if success then fusedSomething (StmNode new_stm) else pure Nothing
-          _ -> error "Illegal result of tryFuseWithAccs called from vFuseNodeT."
+          Just _ -> error "Illegal result of tryFuseWithAccs called from vFuseNodeT."
 --
 vFuseNodeT _ _ _ _ = pure Nothing
 
@@ -551,16 +551,26 @@ runInnerFusionOnContext c@(incoming, node, nodeT, outgoing) = case nodeT of
     lam' <- fst <$> doFusionInLambda lam
     pure (incoming, node, StmNode (Let pat aux (WithAcc inputs lam')), outgoing)
   SoacNode ots pat soac aux -> do
-    let lam = H.lambda soac
-    lam' <- inScopeOf lam $ case soac of
-      H.Stream {} ->
-        dontFuseScans $ fst <$> doFusionInLambda lam
+    soac' <- case soac of
+      H.Stream w inputs accs lam ->
+        H.Stream w inputs accs <$> dontFuseScans (onLambda lam)
+      H.Screma w inputs (ScremaForm lam scans reds) ->
+        H.Screma w inputs
+          <$> ( ScremaForm
+                  <$> doFuseScans (onLambda lam)
+                  <*> mapM onScan scans
+                  <*> mapM onRed reds
+              )
       _ ->
-        doFuseScans $ fst <$> doFusionInLambda lam
-    let nodeT' = SoacNode ots pat (H.setLambda lam' soac) aux
+        H.setLambda <$> doFuseScans (onLambda (H.lambda soac)) <*> pure soac
+    let nodeT' = SoacNode ots pat soac' aux
     pure (incoming, node, nodeT', outgoing)
   _ -> pure c
   where
+    onLambda lam = inScopeOf lam . fmap fst $ doFusionInLambda lam
+    onScan (Scan lam nes) = Scan <$> onLambda lam <*> pure nes
+    onRed (Reduce comm lam nes) = Reduce comm <$> onLambda lam <*> pure nes
+
     doFusionWithDelayed :: Body SOACS -> [(NodeT, [EdgeT])] -> FusionM (Body SOACS)
     doFusionWithDelayed (Body () stms res) extraNodes = inScopeOf stms $ do
       stm_node <- mapM (finalizeNode . fst) extraNodes
