@@ -18,9 +18,10 @@ import Control.Arrow (first)
 import Control.Monad
 import Control.Monad.Reader
 import Control.Monad.State
-import Data.List (find, tails, (\\))
+import Data.List (find, (\\))
 import Data.Map.Strict qualified as M
 import Data.Maybe
+import Futhark.Analysis.HORep.MapNest (MapNest)
 import Futhark.Analysis.HORep.MapNest qualified as MapNest
 import Futhark.Analysis.HORep.SOAC qualified as SOAC
 import Futhark.Construct
@@ -66,8 +67,6 @@ liftMaybe (Just x) = pure x
 
 type SOAC = SOAC.SOAC SOACS
 
-type MapNest = MapNest.MapNest SOACS
-
 inputToOutput :: SOAC.Input -> Maybe (SOAC.ArrayTransform, SOAC.Input)
 inputToOutput (SOAC.Input ts ia iat) =
   case SOAC.viewf ts of
@@ -409,6 +408,10 @@ fuseSOACwithKer mode unfus_set outVars soac_p ker = do
       fail "Cannot fuse a scatter with anything else than a scatter or a map"
     (_, SOAC.Scatter {}, _) ->
       fail "Cannot fuse a scatter with anything else than a scatter or a map"
+    (_, SOAC.Hist {}, _) ->
+      fail "Cannot fuse a Hist with anything else than a Hist or a Map"
+    (SOAC.Hist {}, _, _) ->
+      fail "Cannot fuse a Hist with anything else than a Hist or a Map"
     ----------------------------
     -- Stream-Stream Fusions: --
     ----------------------------
@@ -559,7 +562,7 @@ iswim ::
   TryFusion (SOAC, SOAC.ArrayTransforms)
 iswim _ (SOAC.Screma w arrs form) ots
   | Just [Futhark.Scan scan_fun nes] <- Futhark.isScanSOAC form,
-    Just (map_pat, map_cs, map_w, map_fun) <- rwimPossible scan_fun,
+    Just (map_pat, map_aux, map_w, map_fun) <- rwimPossible scan_fun,
     Just nes_names <- mapM subExpVar nes = do
       let nes_idents = zipWith Ident nes_names $ lambdaReturnType scan_fun
           map_nes = map SOAC.identInput nes_idents
@@ -595,7 +598,7 @@ iswim _ (SOAC.Screma w arrs form) ots
 
       pure
         ( SOAC.Screma map_w map_arrs' (mapSOAC map_fun'),
-          ots SOAC.|> SOAC.Rearrange map_cs perm
+          ots SOAC.|> SOAC.Rearrange map_aux perm
         )
 iswim _ _ _ =
   fail "ISWIM does not apply."
@@ -767,52 +770,19 @@ fixupInputs inpIds inps =
       | otherwise = Nothing
 
 pullReshape :: SOAC -> SOAC.ArrayTransforms -> TryFusion (SOAC, SOAC.ArrayTransforms)
-pullReshape (SOAC.Screma _ inps form) ots
-  | Just maplam <- Futhark.isMapSOAC form,
-    SOAC.Reshape cs k shape SOAC.:< ots' <- SOAC.viewf ots,
-    all primType $ lambdaReturnType maplam = do
-      let mapw' = case reverse $ shapeDims shape of
-            [] -> intConst Int64 0
-            d : _ -> d
-          trInput inp
-            | arrayRank (SOAC.inputType inp) == 1 =
-                SOAC.addTransform (SOAC.Reshape cs k shape) inp
-            | otherwise =
-                SOAC.addTransform (SOAC.ReshapeOuter cs k shape) inp
-          inputs' = map trInput inps
-          inputTypes = map SOAC.inputType inputs'
-
-      let outersoac ::
-            ([SOAC.Input] -> SOAC) ->
-            (SubExp, [SubExp]) ->
-            TryFusion ([SOAC.Input] -> SOAC)
-          outersoac inner (w, outershape) = do
-            let addDims t = arrayOf t (Shape outershape) NoUniqueness
-                retTypes = map addDims $ lambdaReturnType maplam
-
-            ps <- forM inputTypes $ \inpt ->
-              newParam "pullReshape_param" $
-                stripArray (length shape - length outershape) inpt
-
-            inner_body <-
-              runBodyBuilder $
-                varsRes
-                  <$> (letTupExp "x" <=< SOAC.toExp $ inner $ map (SOAC.identInput . paramIdent) ps)
-            let inner_fun =
-                  Lambda
-                    { lambdaParams = ps,
-                      lambdaReturnType = retTypes,
-                      lambdaBody = inner_body
-                    }
-            pure $ flip (SOAC.Screma w) $ Futhark.mapSOAC inner_fun
-
-      op' <-
-        foldM outersoac (flip (SOAC.Screma mapw') $ Futhark.mapSOAC maplam) $
-          zip (drop 1 $ reverse $ shapeDims shape) $
-            drop 1 . reverse . drop 1 . tails $
-              shapeDims shape
-      pure (op' inputs', ots')
-pullReshape _ _ = fail "Cannot pull reshape"
+pullReshape soac ots = do
+  Just mapnest <- MapNest.fromSOAC soac
+  SOAC.Reshape cs newshape SOAC.:< ots' <- pure $ SOAC.viewf ots
+  -- This handles only the easy case where the underlying lambda is
+  -- scalar. The more complicated cases could also be handled, but
+  -- requires more tricky checks.
+  guard $
+    all
+      ((== MapNest.depth mapnest) . arrayRank)
+      (MapNest.typeOf mapnest)
+  mapnest' <- MapNest.reshape cs (newShape newshape) mapnest
+  soac' <- MapNest.toSOAC mapnest'
+  pure (soac', ots')
 
 -- Tie it all together in exposeInputs (for making inputs to a
 -- consumer available) and pullOutputTransforms (for moving
@@ -825,6 +795,7 @@ exposeInputs ::
 exposeInputs inpIds ker =
   (exposeInputs' =<< pushRearrange')
     <|> (exposeInputs' =<< pullRearrange')
+    <|> (exposeInputs' =<< pullReshape')
     <|> (exposeInputs' =<< pullIndex')
     <|> exposeInputs' ker
   where
@@ -848,6 +819,16 @@ exposeInputs inpIds ker =
             fsOutputTransform = SOAC.noTransforms
           }
 
+    pullReshape' = do
+      (soac', ot') <- pullReshape (fsSOAC ker) ot
+      unless (SOAC.nullTransforms ot') $
+        fail "pullReshape was not enough"
+      pure
+        ker
+          { fsSOAC = soac',
+            fsOutputTransform = SOAC.noTransforms
+          }
+
     pullIndex' = do
       (soac', ot') <- pullIndex (fsSOAC ker) ot
       unless (SOAC.nullTransforms ot') $
