@@ -13,6 +13,7 @@ module Futhark.Optimise.TileLoops.Shared
     varianceInStms,
     isTileableRedomap,
     changeEnv,
+    initialIxFnEnv,
     TileKind (..),
   )
 where
@@ -275,6 +276,16 @@ changeEnv (with_env, ixfn_env) y e = do
   ixfn_env' <- changeIxFnEnv ixfn_env y e
   pure (with_env', ixfn_env')
 
+-- | Construct an initial 'IxFnEnv' where it is assumed that every array
+-- parameter in the scope has a row-major index function.
+initialIxFnEnv :: Scope GPU -> IxFnEnv
+initialIxFnEnv = M.mapMaybe f
+  where
+    f info =
+      case typeOf info of
+        Array _ shape _ -> Just $ LMAD.iota 0 $ map pe64 $ shapeDims shape
+        _ -> Nothing
+
 changeWithEnv :: WithEnv -> Exp GPU -> TileM WithEnv
 changeWithEnv with_env (WithAcc accum_decs inner_lam) = do
   let bindings = map mapfun accum_decs
@@ -302,11 +313,13 @@ composeIxfuns env y x ixf_fun =
         _ -> env
 
 changeIxFnEnv :: IxFnEnv -> VName -> Exp GPU -> TileM IxFnEnv
-changeIxFnEnv env y (BasicOp (Reshape ReshapeArbitrary shp_chg x)) =
-  composeIxfuns env y x (`LMAD.reshape` fmap ExpMem.pe64 (shapeDims shp_chg))
-changeIxFnEnv env y (BasicOp (Reshape ReshapeCoerce shp_chg x)) =
-  composeIxfuns env y x (Just . (`LMAD.coerce` fmap ExpMem.pe64 (shapeDims shp_chg)))
-changeIxFnEnv env y (BasicOp (Manifest perm x)) = do
+changeIxFnEnv env y (BasicOp (Reshape x shp_chg)) =
+  case reshapeKind shp_chg of
+    ReshapeCoerce ->
+      composeIxfuns env y x (Just . (`LMAD.coerce` fmap ExpMem.pe64 (shapeDims $ newShape shp_chg)))
+    ReshapeArbitrary ->
+      composeIxfuns env y x (`LMAD.reshape` fmap ExpMem.pe64 (shapeDims $ newShape shp_chg))
+changeIxFnEnv env y (BasicOp (Manifest x perm)) = do
   tp <- lookupType x
   case tp of
     Array _ptp shp _u -> do
@@ -314,7 +327,7 @@ changeIxFnEnv env y (BasicOp (Manifest perm x)) = do
       let ixfn = LMAD.permute (LMAD.iota 0 shp') perm
       pure $ M.insert y ixfn env
     _ -> error "In TileLoops/Shared.hs, changeIxFnEnv: manifest applied to a non-array!"
-changeIxFnEnv env y (BasicOp (Rearrange perm x)) =
+changeIxFnEnv env y (BasicOp (Rearrange x perm)) =
   composeIxfuns env y x (Just . (`LMAD.permute` perm))
 changeIxFnEnv env y (BasicOp (Index x slc)) =
   composeIxfuns env y x (Just . (`LMAD.slice` Slice (map (fmap ExpMem.pe64) $ unSlice slc)))
