@@ -334,13 +334,13 @@ distributeMapBodyStms ::
 distributeMapBodyStms orig_acc = distribute <=< onStms orig_acc . stmsToList
   where
     onStms acc [] = pure acc
-    onStms acc (Let pat (StmAux cs _ _) (Op (Stream w arrs accs lam)) : stms) = do
+    onStms acc (Let pat aux (Op (Stream w arrs accs lam)) : stms) = do
       types <- asksScope scopeForSOACs
       stream_stms <-
         snd <$> runBuilderT (sequentialStreamWholeArray pat w accs lam arrs) types
       stream_stms' <-
         runReaderT (copyPropagateInStms simpleSOACS types stream_stms) types
-      onStms acc $ stmsToList (fmap (certify cs) stream_stms') ++ stms
+      onStms acc $ stmsToList (fmap (certify (stmAuxCerts aux)) stream_stms') ++ stms
     onStms acc (stm : stms) =
       -- It is important that stm is in scope if 'maybeDistributeStm'
       -- wants to distribute, even if this causes the slightly silly
@@ -463,7 +463,7 @@ maybeDistributeStm (Let pat aux (Op (Screma w arrs form))) acc
       distributeMapBodyStms acc stms
 
 -- Parallelise segmented scatters.
-maybeDistributeStm stm@(Let pat (StmAux cs _ _) (Op (Scatter w ivs as lam))) acc =
+maybeDistributeStm stm@(Let pat aux (Op (Scatter w ivs as lam))) acc =
   distributeSingleStm acc stm >>= \case
     Just (kernels, res, nest, acc')
       | Just (perm, pat_unused) <- permutationAndMissing pat res ->
@@ -471,12 +471,12 @@ maybeDistributeStm stm@(Let pat (StmAux cs _ _) (Op (Scatter w ivs as lam))) acc
             nest' <- expandKernelNest pat_unused nest
             lam' <- soacsLambda lam
             addPostStms kernels
-            postStm =<< segmentedScatterKernel nest' perm pat cs w lam' ivs as
+            postStm =<< segmentedScatterKernel nest' perm pat (stmAuxCerts aux) w lam' ivs as
             pure acc'
     _ ->
       addStmToAcc stm acc
 -- Parallelise segmented Hist.
-maybeDistributeStm stm@(Let pat (StmAux cs _ _) (Op (Hist w as ops lam))) acc =
+maybeDistributeStm stm@(Let pat aux (Op (Hist w as ops lam))) acc =
   distributeSingleStm acc stm >>= \case
     Just (kernels, res, nest, acc')
       | Just (perm, pat_unused) <- permutationAndMissing pat res ->
@@ -484,7 +484,7 @@ maybeDistributeStm stm@(Let pat (StmAux cs _ _) (Op (Hist w as ops lam))) acc =
             lam' <- soacsLambda lam
             nest' <- expandKernelNest pat_unused nest
             addPostStms kernels
-            postStm =<< segmentedHistKernel nest' perm cs w ops lam' as
+            postStm =<< segmentedHistKernel nest' perm (stmAuxCerts aux) w ops lam' as
             pure acc'
     _ ->
       addStmToAcc stm acc
@@ -507,7 +507,7 @@ maybeDistributeStm stm@(Let (Pat [pe]) aux (BasicOp (Index arr slice))) acc
 --
 -- If the scan cannot be distributed by itself, it will be
 -- sequentialised in the default case for this function.
-maybeDistributeStm stm@(Let pat (StmAux cs _ _) (Op (Screma w arrs form))) acc
+maybeDistributeStm stm@(Let pat aux (Op (Screma w arrs form))) acc
   | Just (scans, map_lam) <- isScanomapSOAC form,
     Scan lam nes <- singleScan scans =
       distributeSingleStm acc stm >>= \case
@@ -519,7 +519,7 @@ maybeDistributeStm stm@(Let pat (StmAux cs _ _) (Op (Screma w arrs form))) acc
                 nest' <- expandKernelNest pat_unused nest
                 map_lam' <- soacsLambda map_lam
                 localScope (typeEnvFromDistAcc acc') $
-                  segmentedScanomapKernel nest' perm cs w lam map_lam' nes arrs
+                  segmentedScanomapKernel nest' perm (stmAuxCerts aux) w lam map_lam' nes arrs
                     >>= kernelOrNot mempty stm acc kernels acc'
         _ ->
           addStmToAcc stm acc
@@ -536,7 +536,7 @@ maybeDistributeStm (Let pat aux (Op (Screma w arrs form))) acc
 --
 -- If the reduction cannot be distributed by itself, it will be
 -- sequentialised in the default case for this function.
-maybeDistributeStm stm@(Let pat (StmAux cs _ _) (Op (Screma w arrs form))) acc
+maybeDistributeStm stm@(Let pat aux (Op (Screma w arrs form))) acc
   | Just (reds, map_lam) <- isRedomapSOAC form,
     Reduce comm lam nes <- singleReduce reds =
       distributeSingleStm acc stm >>= \case
@@ -554,15 +554,15 @@ maybeDistributeStm stm@(Let pat (StmAux cs _ _) (Op (Screma w arrs form))) acc
                       | commutativeLambda lam = Commutative
                       | otherwise = comm
 
-                regularSegmentedRedomapKernel nest' perm cs w comm' lam' map_lam' nes arrs
+                regularSegmentedRedomapKernel nest' perm (stmAuxCerts aux) w comm' lam' map_lam' nes arrs
                   >>= kernelOrNot mempty stm acc kernels acc'
         _ ->
           addStmToAcc stm acc
-maybeDistributeStm (Let pat (StmAux cs _ _) (Op (Screma w arrs form))) acc = do
+maybeDistributeStm (Let pat aux (Op (Screma w arrs form))) acc = do
   -- This Screma is too complicated for us to immediately do
   -- anything, so split it up and try again.
   scope <- asksScope scopeForSOACs
-  distributeMapBodyStms acc . fmap (certify cs) . snd
+  distributeMapBodyStms acc . fmap (certify (stmAuxCerts aux)) . snd
     =<< runBuilderT (dissectScrema pat w form arrs) scope
 maybeDistributeStm stm@(Let _ aux (BasicOp (Replicate shape (Var stm_arr)))) acc = do
   distributeSingleUnaryStm acc stm stm_arr $ \nest outerpat arr ->
@@ -576,14 +576,14 @@ maybeDistributeStm stm@(Let _ aux (BasicOp (Replicate shape (Var stm_arr)))) acc
         -- Move the to-be-replicated dimensions outermost.
         arr_tr <-
           letExp (baseString arr <> "_tr") . BasicOp $
-            Rearrange ([nest_r .. arr_r - 1] ++ [0 .. nest_r - 1]) arr
+            Rearrange arr ([nest_r .. arr_r - 1] ++ [0 .. nest_r - 1])
         -- Replicate the now-outermost dimensions appropriately.
         arr_tr_rep <-
           letExp (baseString arr <> "_tr_rep") . BasicOp $
             Replicate shape (Var arr_tr)
         -- Move the replicated dimensions back where they belong.
         letBind outerpat . BasicOp $
-          Rearrange ([res_r - nest_r .. res_r - 1] ++ [0 .. res_r - nest_r - 1]) arr_tr_rep
+          Rearrange arr_tr_rep ([res_r - nest_r .. res_r - 1] ++ [0 .. res_r - nest_r - 1])
 maybeDistributeStm stm@(Let _ aux (BasicOp (Replicate shape v))) acc = do
   distributeSingleStm acc stm >>= \case
     Just (kernels, _, nest, acc')
@@ -602,7 +602,7 @@ maybeDistributeStm stm@(Let (Pat [pe]) aux (BasicOp (Opaque _ (Var stm_arr)))) a
   | not $ primType $ typeOf pe =
       distributeSingleUnaryStm acc stm stm_arr $ \_ outerpat arr ->
         pure $ oneStm $ Let outerpat aux $ BasicOp $ Replicate mempty $ Var arr
-maybeDistributeStm stm@(Let _ aux (BasicOp (Rearrange perm stm_arr))) acc =
+maybeDistributeStm stm@(Let _ aux (BasicOp (Rearrange stm_arr perm))) acc =
   distributeSingleUnaryStm acc stm stm_arr $ \nest outerpat arr -> do
     let r = length (snd nest) + 1
         perm' = [0 .. r - 1] ++ map (+ r) perm
@@ -613,12 +613,13 @@ maybeDistributeStm stm@(Let _ aux (BasicOp (Rearrange perm stm_arr))) acc =
     pure $
       stmsFromList
         [ Let (Pat [PatElem arr' arr_t]) aux $ BasicOp $ Replicate mempty $ Var arr,
-          Let outerpat aux $ BasicOp $ Rearrange perm' arr'
+          Let outerpat aux $ BasicOp $ Rearrange arr' perm'
         ]
-maybeDistributeStm stm@(Let _ aux (BasicOp (Reshape k reshape stm_arr))) acc =
+maybeDistributeStm stm@(Let _ aux (BasicOp (Reshape stm_arr reshape))) acc =
   distributeSingleUnaryStm acc stm stm_arr $ \nest outerpat arr -> do
-    let reshape' = Shape (kernelNestWidths nest) <> reshape
-    pure $ oneStm $ Let outerpat aux $ BasicOp $ Reshape k reshape' arr
+    let outer = Shape (kernelNestWidths nest)
+        reshape' = reshapeCoerce outer <> newshapeInner outer reshape
+    pure $ oneStm $ Let outerpat aux $ BasicOp $ Reshape arr reshape'
 maybeDistributeStm stm@(Let pat aux (BasicOp (Update _ arr slice (Var v)))) acc
   | not $ null $ sliceDims slice =
       distributeSingleStm acc stm >>= \case
@@ -770,7 +771,7 @@ segmentedScatterKernel nest perm scatter_pat cs scatter_w lam ivs dests = do
   -- KernelNest we produce here is technically not sensible, but it's
   -- good enough for flatKernel to work.
   let nesting =
-        MapNesting scatter_pat (StmAux cs mempty ()) scatter_w $ zip (lambdaParams lam) ivs
+        MapNesting scatter_pat (StmAux cs mempty mempty ()) scatter_w $ zip (lambdaParams lam) ivs
       nest' =
         pushInnerKernelNesting (scatter_pat, bodyResult $ lambdaBody lam) nesting nest
   (ispace, kernel_inps) <- flatKernel nest'
@@ -795,7 +796,9 @@ segmentedScatterKernel nest perm scatter_pat cs scatter_w lam ivs dests = do
 
   dest_arrs_ts <- mapM (lookupType . kernelInputArray) dest_arrs
 
-  let k_body = KernelBody () k_body_stms (zipWith (inPlaceReturn ispace) dest_arrs_ts grouped)
+  let k_body =
+        KernelBody () k_body_stms $
+          zipWith (inPlaceReturn ispace) dest_arrs_ts grouped
       -- Remove unused kernel inputs, since some of these might
       -- reference the array we are scattering into.
       kernel_inps' =
@@ -814,15 +817,17 @@ segmentedScatterKernel nest perm scatter_pat cs scatter_w lam ivs dests = do
     bad = error "Ill-typed nested scatter encountered."
 
     inPlaceReturn ispace arr_t (_, inp, is_vs) =
-      WriteReturns
-        ( foldMap (foldMap resCerts . fst) is_vs
-            <> foldMap (resCerts . snd) is_vs
-        )
-        (kernelInputArray inp)
-        [ (fullSlice arr_t $ map DimFix $ map Var (init gtids) ++ map resSubExp is, resSubExp v)
-          | (is, v) <- is_vs
-        ]
+      WriteReturns write_cs (kernelInputArray inp) $ do
+        (is, v) <- is_vs
+        pure
+          ( fullSlice arr_t . map DimFix $
+              map Var (init gtids) ++ map resSubExp is,
+            resSubExp v
+          )
       where
+        write_cs =
+          foldMap (foldMap resCerts . fst) is_vs
+            <> foldMap (resCerts . snd) is_vs
         (gtids, _ws) = unzip ispace
 
 segmentedUpdateKernel ::
