@@ -6,6 +6,7 @@ module Language.Futhark.Pretty
   ( prettyString,
     prettyTuple,
     leadingOperator,
+    symbolName,
     IsName (..),
     prettyNameString,
     Annot (..),
@@ -37,7 +38,7 @@ import Prelude
 -- with the tag.  To avoid erroneously using the 'Pretty' instance for
 -- VNames, we in fact only define it inside the modules for the core
 -- language (as an orphan instance).
-class IsName v where
+class (Eq v) => IsName v where
   prettyName :: v -> Doc a
   toName :: v -> Name
 
@@ -204,7 +205,7 @@ hasArrayLit ArrayLit {} = True
 hasArrayLit (TupLit es2 _) = any hasArrayLit es2
 hasArrayLit _ = False
 
-instance (Eq vn, IsName vn, Annot f) => Pretty (DimIndexBase f vn) where
+instance (IsName vn, Annot f) => Pretty (DimIndexBase f vn) where
   pretty (DimFix e) = pretty e
   pretty (DimSlice i j (Just s)) =
     maybe mempty pretty i
@@ -223,12 +224,12 @@ instance (Eq vn, IsName vn, Annot f) => Pretty (DimIndexBase f vn) where
 instance (IsName vn) => Pretty (SizeBinder vn) where
   pretty (SizeBinder v _) = brackets $ prettyName v
 
-letBody :: (Eq vn, IsName vn, Annot f) => ExpBase f vn -> Doc a
+letBody :: (IsName vn, Annot f) => ExpBase f vn -> Doc a
 letBody body@(AppExp LetPat {} _) = pretty body
 letBody body@(AppExp LetFun {} _) = pretty body
 letBody body = "in" <+> align (pretty body)
 
-prettyAppExp :: (Eq vn, IsName vn, Annot f) => Int -> AppExpBase f vn -> Doc a
+prettyAppExp :: (IsName vn, Annot f) => Int -> AppExpBase f vn -> Doc a
 prettyAppExp p (BinOp (bop, _) _ (x, _) (y, _) _) = prettyBinOp p bop x y
 prettyAppExp _ (Match e cs _) = "match" <+> pretty e </> (stack . map pretty) (NE.toList cs)
 prettyAppExp _ (Loop sizeparams pat initexp form loopbody _) =
@@ -308,7 +309,7 @@ prettyAppExp p (Apply f args _) =
     prettyExp 0 f
       <+> hsep (map (prettyExp 10 . snd) $ NE.toList args)
 
-instance (Eq vn, IsName vn, Annot f) => Pretty (AppExpBase f vn) where
+instance (IsName vn, Annot f) => Pretty (AppExpBase f vn) where
   pretty = prettyAppExp (-1)
 
 prettyInst :: (Annot f, Pretty t) => f t -> Doc a
@@ -322,17 +323,18 @@ prettyInst t =
 prettyAttr :: (Pretty a) => a -> Doc ann
 prettyAttr attr = "#[" <> pretty attr <> "]"
 
-operatorName :: Name -> Bool
-operatorName = (`elem` opchars) . T.head . nameToText
+-- | Does this name correspond to a symbol rather than an identifier?
+symbolName :: Name -> Bool
+symbolName = (`elem` opchars) . T.head . nameToText
   where
     opchars :: String
     opchars = "+-*/%=!><|&^."
 
-prettyExp :: (Eq vn, IsName vn, Annot f) => Int -> ExpBase f vn -> Doc a
+prettyExp :: (IsName vn, Annot f) => Int -> ExpBase f vn -> Doc a
 prettyExp _ (Var name t _)
   -- The first case occurs only for programs that have been normalised
   -- by the compiler.
-  | operatorName (toName (qualLeaf name)) = parens $ pretty name <> prettyInst t
+  | symbolName (toName (qualLeaf name)) = parens $ pretty name <> prettyInst t
   | otherwise = pretty name <> prettyInst t
 prettyExp _ (Hole t _) = "???" <> prettyInst t
 prettyExp _ (Parens e _) = align $ parens $ pretty e
@@ -359,7 +361,7 @@ prettyExp _ (ArrayLit es t _) =
   brackets (commasep $ map pretty es) <> prettyInst t
 prettyExp _ (StringLit s _) =
   pretty $ show $ map (chr . fromIntegral) s
-prettyExp _ (Project k e _ _) = pretty e <> "." <> pretty k
+prettyExp _ (Project k e _ _) = prettyExp 11 e <> "." <> pretty k
 prettyExp _ (Negate e _) = "-" <> pretty e
 prettyExp _ (Not e _) = "!" <> pretty e
 prettyExp _ (Update src idxs ve _) =
@@ -409,7 +411,7 @@ prettyExp i (AppExp e res)
         <> parens (pretty t <> "," <+> brackets (commasep $ map prettyName ext))
   | otherwise = prettyAppExp i e
 
-instance (Eq vn, IsName vn, Annot f) => Pretty (ExpBase f vn) where
+instance (IsName vn, Annot f) => Pretty (ExpBase f vn) where
   pretty = prettyExp (-1)
 
 instance (IsName vn) => Pretty (AttrAtom vn) where
@@ -420,18 +422,18 @@ instance (IsName vn) => Pretty (AttrInfo vn) where
   pretty (AttrAtom attr _) = pretty attr
   pretty (AttrComp f attrs _) = pretty f <> parens (commasep $ map pretty attrs)
 
-instance (Eq vn, IsName vn, Annot f) => Pretty (FieldBase f vn) where
+instance (IsName vn, Annot f) => Pretty (FieldBase f vn) where
   pretty (RecordFieldExplicit (L _ name) e _) = pretty name <> equals <> pretty e
   pretty (RecordFieldImplicit (L _ name) _ _) = prettyName name
 
-instance (Eq vn, IsName vn, Annot f) => Pretty (CaseBase f vn) where
+instance (IsName vn, Annot f) => Pretty (CaseBase f vn) where
   pretty (CasePat p e _) = "case" <+> pretty p <+> "->" </> indent 2 (pretty e)
 
-instance (Eq vn, IsName vn, Annot f) => Pretty (LoopInitBase f vn) where
+instance (IsName vn, Annot f) => Pretty (LoopInitBase f vn) where
   pretty (LoopInitImplicit e) = maybe "_" pretty $ unAnnot e
   pretty (LoopInitExplicit e) = pretty e
 
-instance (Eq vn, IsName vn, Annot f) => Pretty (LoopFormBase f vn) where
+instance (IsName vn, Annot f) => Pretty (LoopFormBase f vn) where
   pretty (For i ubound) =
     "for" <+> pretty i <+> "<" <+> align (pretty ubound)
   pretty (ForIn x e) =
@@ -444,7 +446,7 @@ instance Pretty PatLit where
   pretty (PatLitFloat f) = pretty f
   pretty (PatLitPrim v) = pretty v
 
-instance (Eq vn, IsName vn, Annot f, Pretty t) => Pretty (PatBase f vn t) where
+instance (IsName vn, Annot f, Pretty t) => Pretty (PatBase f vn t) where
   pretty (PatAscription p t _) = pretty p <> colon <+> align (pretty t)
   pretty (PatParens p _) = parens $ pretty p
   pretty (Id v t _) = case unAnnot t of
@@ -465,10 +467,10 @@ ppAscription :: (Pretty t) => Maybe t -> Doc a
 ppAscription Nothing = mempty
 ppAscription (Just t) = colon <> align (pretty t)
 
-instance (Eq vn, IsName vn, Annot f) => Pretty (ProgBase f vn) where
+instance (IsName vn, Annot f) => Pretty (ProgBase f vn) where
   pretty = stack . punctuate line . map pretty . progDecs
 
-instance (Eq vn, IsName vn, Annot f) => Pretty (DecBase f vn) where
+instance (IsName vn, Annot f) => Pretty (DecBase f vn) where
   pretty (ValDec dec) = pretty dec
   pretty (TypeDec dec) = pretty dec
   pretty (ModTypeDec sig) = pretty sig
@@ -477,7 +479,7 @@ instance (Eq vn, IsName vn, Annot f) => Pretty (DecBase f vn) where
   pretty (LocalDec dec _) = "local" <+> pretty dec
   pretty (ImportDec x _ _) = "import" <+> pretty x
 
-prettyModExp :: (Eq vn, IsName vn, Annot f) => Int -> ModExpBase f vn -> Doc a
+prettyModExp :: (IsName vn, Annot f) => Int -> ModExpBase f vn -> Doc a
 prettyModExp _ (ModVar v _) =
   pretty v
 prettyModExp _ (ModParens e _) =
@@ -502,7 +504,7 @@ prettyModExp p (ModLambda param maybe_sig body _) =
       Nothing -> mempty
       Just (sig, _) -> colon <+> pretty sig
 
-instance (Eq vn, IsName vn, Annot f) => Pretty (ModExpBase f vn) where
+instance (IsName vn, Annot f) => Pretty (ModExpBase f vn) where
   pretty = prettyModExp (-1)
 
 instance Pretty Liftedness where
@@ -510,7 +512,7 @@ instance Pretty Liftedness where
   pretty SizeLifted = "~"
   pretty Lifted = "^"
 
-instance (Eq vn, IsName vn, Annot f) => Pretty (TypeBindBase f vn) where
+instance (IsName vn, Annot f) => Pretty (TypeBindBase f vn) where
   pretty (TypeBind name l params te rt _ _) =
     "type"
       <> pretty l
@@ -518,11 +520,11 @@ instance (Eq vn, IsName vn, Annot f) => Pretty (TypeBindBase f vn) where
         <+> equals
         <+> maybe (pretty te) pretty (unAnnot rt)
 
-instance (Eq vn, IsName vn) => Pretty (TypeParamBase vn) where
+instance (IsName vn) => Pretty (TypeParamBase vn) where
   pretty (TypeParamDim name _) = brackets $ prettyName name
   pretty (TypeParamType l name _) = "'" <> pretty l <> prettyName name
 
-instance (Eq vn, IsName vn, Annot f) => Pretty (ValBindBase f vn) where
+instance (IsName vn, Annot f) => Pretty (ValBindBase f vn) where
   pretty (ValBind entry name retdecl rettype tparams args body _ attrs _) =
     mconcat (map ((<> line) . prettyAttr) attrs)
       <> fun
@@ -544,18 +546,23 @@ instance (Eq vn, IsName vn, Annot f) => Pretty (ValBindBase f vn) where
         Just rettype' -> [colon <+> align rettype']
         Nothing -> mempty
 
-instance (Eq vn, IsName vn, Annot f) => Pretty (SpecBase f vn) where
+instance (IsName vn, Annot f) => Pretty (SpecBase f vn) where
   pretty (TypeAbbrSpec tpsig) = pretty tpsig
   pretty (TypeSpec l name ps _ _) =
     "type" <> pretty l <+> hsep (prettyName name : map pretty ps)
   pretty (ValSpec name tparams vtype _ _ _) =
-    "val" <+> hsep (prettyName name : map pretty tparams) <> colon <+> pretty vtype
+    "val" <+> hsep (name' : map pretty tparams) <> colon <+> pretty vtype
+    where
+      name' =
+        if symbolName $ toName name
+          then parens $ prettyName name
+          else prettyName name
   pretty (ModSpec name sig _ _) =
     "module" <+> prettyName name <> colon <+> pretty sig
   pretty (IncludeSpec e _) =
     "include" <+> pretty e
 
-instance (Eq vn, IsName vn, Annot f) => Pretty (ModTypeExpBase f vn) where
+instance (IsName vn, Annot f) => Pretty (ModTypeExpBase f vn) where
   pretty (ModTypeVar v _ _) = pretty v
   pretty (ModTypeParens e _) = parens $ pretty e
   pretty (ModTypeSpecs ss _) = nestedBlock "{" "}" (stack $ punctuate line $ map pretty ss)
@@ -566,15 +573,15 @@ instance (Eq vn, IsName vn, Annot f) => Pretty (ModTypeExpBase f vn) where
   pretty (ModTypeArrow Nothing e1 e2 _) =
     pretty e1 <+> "->" <+> pretty e2
 
-instance (Eq vn, IsName vn, Annot f) => Pretty (ModTypeBindBase f vn) where
+instance (IsName vn, Annot f) => Pretty (ModTypeBindBase f vn) where
   pretty (ModTypeBind name e _ _) =
     "module type" <+> prettyName name <+> equals <+> pretty e
 
-instance (Eq vn, IsName vn, Annot f) => Pretty (ModParamBase f vn) where
+instance (IsName vn, Annot f) => Pretty (ModParamBase f vn) where
   pretty (ModParam pname psig _ _) =
     parens (prettyName pname <> colon <+> pretty psig)
 
-instance (Eq vn, IsName vn, Annot f) => Pretty (ModBindBase f vn) where
+instance (IsName vn, Annot f) => Pretty (ModBindBase f vn) where
   pretty (ModBind name ps sig e _ _) =
     "module" <+> hsep (prettyName name : map pretty ps) <> sig' <> " =" <+> pretty e
     where
@@ -591,7 +598,7 @@ ppBinOp bop =
     leading = leadingOperator $ toName $ qualLeaf bop
 
 prettyBinOp ::
-  (Eq vn, IsName vn, Annot f) =>
+  (IsName vn, Annot f) =>
   Int ->
   QualName vn ->
   ExpBase f vn ->
