@@ -566,6 +566,10 @@ data ConvOp
     UIToFP IntType FloatType
   | -- | Convert a signed integer to a floating-point value.
     SIToFP IntType FloatType
+  | -- | Convert floating point number to its bitwise representation.
+    FPToBits FloatType
+  | -- | Convert bitwise representation to a floating point number.
+    BitsToFP FloatType
   | -- | Convert an integer to a boolean value.  Zero
     -- becomes false; anything else is true.
     IToB IntType
@@ -653,6 +657,8 @@ allConvOps =
       FPToSI <$> allFloatTypes <*> allIntTypes,
       UIToFP <$> allIntTypes <*> allFloatTypes,
       SIToFP <$> allIntTypes <*> allFloatTypes,
+      FPToBits <$> allFloatTypes,
+      BitsToFP <$> allFloatTypes,
       IToB <$> allIntTypes,
       BToI <$> allIntTypes,
       FToB <$> allFloatTypes,
@@ -930,12 +936,30 @@ doConvOp (FPToUI _ to) (FloatValue v) = Just $ IntValue $ doFPToUI v to
 doConvOp (FPToSI _ to) (FloatValue v) = Just $ IntValue $ doFPToSI v to
 doConvOp (UIToFP _ to) (IntValue v) = Just $ FloatValue $ doUIToFP v to
 doConvOp (SIToFP _ to) (IntValue v) = Just $ FloatValue $ doSIToFP v to
+doConvOp FPToBits {} (FloatValue (Float16Value x)) =
+  Just $ IntValue $ Int16Value $ fromIntegral $ halfToWord x
+doConvOp FPToBits {} (FloatValue (Float32Value x)) =
+  Just $ IntValue $ Int32Value $ fromIntegral $ floatToWord x
+doConvOp FPToBits {} (FloatValue (Float64Value x)) =
+  Just $ IntValue $ Int64Value $ fromIntegral $ doubleToWord x
+doConvOp BitsToFP {} (IntValue (Int16Value x)) =
+  Just $ FloatValue $ Float16Value $ wordToHalf $ fromIntegral x
+doConvOp BitsToFP {} (IntValue (Int32Value x)) =
+  Just $ FloatValue $ Float32Value $ wordToFloat $ fromIntegral x
+doConvOp BitsToFP {} (IntValue (Int64Value x)) =
+  Just $ FloatValue $ Float64Value $ wordToDouble $ fromIntegral x
 doConvOp (IToB _) (IntValue v) = Just $ BoolValue $ intToInt64 v /= 0
 doConvOp (BToI to) (BoolValue v) = Just $ IntValue $ intValue to $ if v then 1 else 0 :: Int
 doConvOp (FToB _) (FloatValue v) = Just $ BoolValue $ floatToDouble v /= 0
 doConvOp (BToF to) (BoolValue v) = Just $ FloatValue $ floatValue to $ if v then 1 else 0 :: Double
 doConvOp _ _ = Nothing
 
+-- | The integer type with the same size as the given floating point type.
+matchingInt :: FloatType -> IntType
+matchingInt Float16 = Int16
+matchingInt Float32 = Int32
+matchingInt Float64 = Int64
+
 -- | Turn the conversion the other way around.  Note that most
 -- conversions are lossy, so there is no guarantee the value will
 -- round-trip.
@@ -947,6 +971,8 @@ flipConvOp (FPToUI from to) = UIToFP to from
 flipConvOp (FPToSI from to) = SIToFP to from
 flipConvOp (UIToFP from to) = FPToSI to from
 flipConvOp (SIToFP from to) = FPToSI to from
+flipConvOp (FPToBits from) = BitsToFP from
+flipConvOp (BitsToFP to) = FPToBits to
 flipConvOp (IToB from) = BToI from
 flipConvOp (BToI to) = IToB to
 flipConvOp (FToB from) = BToF from
@@ -1161,6 +1187,8 @@ convOpType (FPToUI from to) = (FloatType from, IntType to)
 convOpType (FPToSI from to) = (FloatType from, IntType to)
 convOpType (UIToFP from to) = (IntType from, FloatType to)
 convOpType (SIToFP from to) = (IntType from, FloatType to)
+convOpType (FPToBits from) = (FloatType from, IntType $ matchingInt from)
+convOpType (BitsToFP to) = (IntType $ matchingInt to, FloatType to)
 convOpType (IToB from) = (IntType from, Bool)
 convOpType (BToI to) = (Bool, IntType to)
 convOpType (FToB from) = (FloatType from, Bool)
@@ -1210,6 +1238,10 @@ primFuns =
       f32 "sqrt32" sqrt,
       f64 "sqrt64" sqrt,
       --
+      f16 "rsqrt16" $ recip . sqrt,
+      f32 "rsqrt32" $ recip . sqrt,
+      f64 "rsqrt64" $ recip . sqrt,
+      --
       f16 "cbrt16" $ convFloat . cbrtf . convFloat,
       f32 "cbrt32" cbrtf,
       f64 "cbrt64" cbrt,
@@ -1238,6 +1270,10 @@ primFuns =
       f32 "sin32" sin,
       f64 "sin64" sin,
       --
+      f16 "sinpi16" $ sin . (pi *),
+      f32 "sinpi32" $ sin . (pi *),
+      f64 "sinpi64" $ sin . (pi *),
+      --
       f16 "sinh16" sinh,
       f32 "sinh32" sinh,
       f64 "sinh64" sinh,
@@ -1246,6 +1282,10 @@ primFuns =
       f32 "cos32" cos,
       f64 "cos64" cos,
       --
+      f16 "cospi16" $ cos . (pi *),
+      f32 "cospi32" $ cos . (pi *),
+      f64 "cospi64" $ cos . (pi *),
+      --
       f16 "cosh16" cosh,
       f32 "cosh32" cosh,
       f64 "cosh64" cosh,
@@ -1254,6 +1294,10 @@ primFuns =
       f32 "tan32" tan,
       f64 "tan64" tan,
       --
+      f16 "tanpi16" $ tan . (pi *),
+      f32 "tanpi32" $ tan . (pi *),
+      f64 "tanpi64" $ tan . (pi *),
+      --
       f16 "tanh16" tanh,
       f32 "tanh32" tanh,
       f64 "tanh64" tanh,
@@ -1262,6 +1306,10 @@ primFuns =
       f32 "asin32" asin,
       f64 "asin64" asin,
       --
+      f16 "asinpi16" $ (/ pi) . asin,
+      f32 "asinpi32" $ (/ pi) . asin,
+      f64 "asinpi64" $ (/ pi) . asin,
+      --
       f16 "asinh16" asinh,
       f32 "asinh32" asinh,
       f64 "asinh64" asinh,
@@ -1270,6 +1318,10 @@ primFuns =
       f32 "acos32" acos,
       f64 "acos64" acos,
       --
+      f16 "acospi16" $ (/ pi) . acos,
+      f32 "acospi32" $ (/ pi) . acos,
+      f64 "acospi64" $ (/ pi) . acos,
+      --
       f16 "acosh16" acosh,
       f32 "acosh32" acosh,
       f64 "acosh64" acosh,
@@ -1278,6 +1330,10 @@ primFuns =
       f32 "atan32" atan,
       f64 "atan64" atan,
       --
+      f16 "atanpi16" $ (/ pi) . atan,
+      f32 "atanpi32" $ (/ pi) . atan,
+      f64 "atanpi64" $ (/ pi) . atan,
+      --
       f16 "atanh16" atanh,
       f32 "atanh32" atanh,
       f64 "atanh64" atanh,
@@ -1403,6 +1459,33 @@ primFuns =
             _ -> Nothing
         )
       ),
+      ( "atan2pi_16",
+        ( [FloatType Float16, FloatType Float16],
+          FloatType Float16,
+          \case
+            [FloatValue (Float16Value x), FloatValue (Float16Value y)] ->
+              Just $ FloatValue $ Float16Value $ atan2 x y / pi
+            _ -> Nothing
+        )
+      ),
+      ( "atan2pi_32",
+        ( [FloatType Float32, FloatType Float32],
+          FloatType Float32,
+          \case
+            [FloatValue (Float32Value x), FloatValue (Float32Value y)] ->
+              Just $ FloatValue $ Float32Value $ atan2 x y / pi
+            _ -> Nothing
+        )
+      ),
+      ( "atan2pi_64",
+        ( [FloatType Float64, FloatType Float64],
+          FloatType Float64,
+          \case
+            [FloatValue (Float64Value x), FloatValue (Float64Value y)] ->
+              Just $ FloatValue $ Float64Value $ atan2 x y / pi
+            _ -> Nothing
+        )
+      ),
       --
       ( "hypot16",
         ( [FloatType Float16, FloatType Float16],
@@ -1479,60 +1562,6 @@ primFuns =
             _ -> Nothing
         )
       ),
-      ( "to_bits16",
-        ( [FloatType Float16],
-          IntType Int16,
-          \case
-            [FloatValue (Float16Value x)] ->
-              Just $ IntValue $ Int16Value $ fromIntegral $ halfToWord x
-            _ -> Nothing
-        )
-      ),
-      ( "to_bits32",
-        ( [FloatType Float32],
-          IntType Int32,
-          \case
-            [FloatValue (Float32Value x)] ->
-              Just $ IntValue $ Int32Value $ fromIntegral $ floatToWord x
-            _ -> Nothing
-        )
-      ),
-      ( "to_bits64",
-        ( [FloatType Float64],
-          IntType Int64,
-          \case
-            [FloatValue (Float64Value x)] ->
-              Just $ IntValue $ Int64Value $ fromIntegral $ doubleToWord x
-            _ -> Nothing
-        )
-      ),
-      ( "from_bits16",
-        ( [IntType Int16],
-          FloatType Float16,
-          \case
-            [IntValue (Int16Value x)] ->
-              Just $ FloatValue $ Float16Value $ wordToHalf $ fromIntegral x
-            _ -> Nothing
-        )
-      ),
-      ( "from_bits32",
-        ( [IntType Int32],
-          FloatType Float32,
-          \case
-            [IntValue (Int32Value x)] ->
-              Just $ FloatValue $ Float32Value $ wordToFloat $ fromIntegral x
-            _ -> Nothing
-        )
-      ),
-      ( "from_bits64",
-        ( [IntType Int64],
-          FloatType Float64,
-          \case
-            [IntValue (Int64Value x)] ->
-              Just $ FloatValue $ Float64Value $ wordToDouble $ fromIntegral x
-            _ -> Nothing
-        )
-      ),
       f16_3 "lerp16" (\v0 v1 t -> v0 + (v1 - v0) * max 0 (min 1 t)),
       f32_3 "lerp32" (\v0 v1 t -> v0 + (v1 - v0) * max 0 (min 1 t)),
       f64_3 "lerp64" (\v0 v1 t -> v0 + (v1 - v0) * max 0 (min 1 t)),
@@ -1895,6 +1924,8 @@ convOpFun FPToUI {} = "fptoui"
 convOpFun FPToSI {} = "fptosi"
 convOpFun UIToFP {} = "uitofp"
 convOpFun SIToFP {} = "sitofp"
+convOpFun FPToBits {} = "fptobits"
+convOpFun BitsToFP {} = "bitstofp"
 convOpFun IToB {} = "itob"
 convOpFun BToI {} = "btoi"
 convOpFun FToB {} = "ftob"
