@@ -7,8 +7,6 @@ module Language.Futhark.Prop
     Intrinsic (..),
     intrinsics,
     intrinsicVar,
-    isBuiltin,
-    isBuiltinLoc,
     maxIntrinsicTag,
     namesToPrimTypes,
     qualName,
@@ -135,7 +133,6 @@ import Data.Char
 import Data.Foldable
 import Data.List (genericLength, isPrefixOf, sortOn)
 import Data.List.NonEmpty qualified as NE
-import Data.Loc (Loc (..), posFile)
 import Data.Map.Strict qualified as M
 import Data.Maybe
 import Data.Ord
@@ -147,7 +144,6 @@ import Language.Futhark.Primitive qualified as Primitive
 import Language.Futhark.Syntax
 import Language.Futhark.Traversals
 import Language.Futhark.Tuple
-import System.FilePath (takeDirectory)
 
 -- | The name of the default program entry point (@main@).
 defaultEntryPoint :: Name
@@ -1198,18 +1194,6 @@ intrinsics =
     tupInt64 x =
       tupleRecord $ replicate x $ Scalar $ Prim $ Signed Int64
 
--- | Is this include part of the built-in prelude?
-isBuiltin :: FilePath -> Bool
-isBuiltin = (== "/prelude") . takeDirectory
-
--- | Is the position of this thing builtin as per 'isBuiltin'?  Things
--- without location are considered not built-in.
-isBuiltinLoc :: (Located a) => a -> Bool
-isBuiltinLoc x =
-  case locOf x of
-    NoLoc -> False
-    Loc pos _ -> isBuiltin $ posFile pos
-
 -- | The largest tag used by an intrinsic - this can be used to
 -- determine whether a 'VName' refers to an intrinsic or a user-defined name.
 maxIntrinsicTag :: Int
