@@ -331,25 +331,18 @@ checkModTypeExp (ModTypeWith s (TypeRef tname ps te trloc) loc) = do
   (abs, s_abs, s_env, s') <- checkModTypeExpToEnv s
   resolveTypeParams ps $ \ps' -> do
     (ext, te', te_t, _) <- bindingTypeParams ps' $ checkTypeDecl te
-    unless (null ext) $
-      typeError te' mempty "Anonymous dimensions are not allowed here."
-    (tname', s_abs', s_env') <- refineEnv loc s_abs s_env tname ps' te_t
+    (tname', s_abs', s_env') <-
+      refineEnv loc s_abs s_env tname ps' $ RetType ext te_t
     pure (abs, MTy s_abs' $ ModEnv s_env', ModTypeWith s' (TypeRef tname' ps' te' trloc) loc)
 checkModTypeExp (ModTypeArrow maybe_pname e1 e2 loc) = do
   (e1_abs, MTy s_abs e1_mod, e1') <- checkModTypeExp e1
-  (env_for_e2, maybe_pname') <-
+  (maybe_pname', (e2_abs, e2_mod, e2')) <-
     case maybe_pname of
       Just pname -> bindSpaced1 Term pname loc $ \pname' ->
-        pure
-          ( mempty
-              { envNameMap = M.singleton (Term, pname) $ qualName pname',
-                envModTable = M.singleton pname' e1_mod
-              },
-            Just pname'
-          )
+        localEnv (mempty {envModTable = M.singleton pname' e1_mod}) $
+          (Just pname',) <$> checkModTypeExp e2
       Nothing ->
-        pure (mempty, Nothing)
-  (e2_abs, e2_mod, e2') <- localEnv env_for_e2 $ checkModTypeExp e2
+        (Nothing,) <$> checkModTypeExp e2
   pure
     ( e1_abs <> e2_abs,
       MTy mempty $ ModFun $ FunModType s_abs e1_mod e2_mod,
@@ -644,10 +637,9 @@ checkEntryPoint ::
   SrcLoc ->
   [TypeParam] ->
   [Pat ParamType] ->
-  Maybe (TypeExp Exp VName) ->
   ResRetType ->
   TypeM ()
-checkEntryPoint loc tparams params maybe_tdecl rettype
+checkEntryPoint loc tparams params rettype
   | any isTypeParam tparams =
       typeError loc mempty $
         withIndexLink
@@ -674,16 +666,6 @@ checkEntryPoint loc tparams params maybe_tdecl rettype
         "Entry point size parameter "
           <> pretty p
           <> " only used non-constructively."
-  | p : _ <- filter nastyParameter params =
-      warn p $
-        "Entry point parameter\n"
-          </> indent 2 (pretty p)
-          </> "\nwill have an opaque type, so the entry point will likely not be callable."
-  | nastyReturnType maybe_tdecl rettype_t =
-      warn loc $
-        "Entry point return type\n"
-          </> indent 2 (pretty rettype)
-          </> "\nwill have an opaque type, so the result will likely not be usable."
   | otherwise =
       pure ()
   where
@@ -708,7 +690,7 @@ checkValBind vb = do
 
   let entry' = Info (entryPoint params' maybe_tdecl' rettype) <$ entry
   case entry' of
-    Just _ -> checkEntryPoint loc tparams' params' maybe_tdecl' rettype
+    Just _ -> checkEntryPoint loc tparams' params' rettype
     _ -> pure ()
 
   let vb' = ValBind entry' fname maybe_tdecl' (Info rettype) tparams' params' body' doc attrs' loc
@@ -722,44 +704,6 @@ checkValBind vb = do
       vb'
     )
 
-nastyType :: (Monoid als) => TypeBase dim als -> Bool
-nastyType (Scalar Prim {}) = False
-nastyType t@Array {} = nastyType $ stripArray 1 t
-nastyType _ = True
-
-nastyReturnType :: (Monoid als) => Maybe (TypeExp Exp VName) -> TypeBase dim als -> Bool
-nastyReturnType Nothing (Scalar (Arrow _ _ _ t1 (RetType _ t2))) =
-  nastyType t1 || nastyReturnType Nothing t2
-nastyReturnType (Just (TEArrow _ te1 te2 _)) (Scalar (Arrow _ _ _ t1 (RetType _ t2))) =
-  (not (niceTypeExp te1) && nastyType t1)
-    || nastyReturnType (Just te2) t2
-nastyReturnType (Just te) _
-  | niceTypeExp te = False
-nastyReturnType te t
-  | Just ts <- isTupleRecord t =
-      case te of
-        Just (TETuple tes _) -> or $ zipWith nastyType' (map Just tes) ts
-        _ -> any nastyType ts
-  | otherwise = nastyType' te t
-  where
-    nastyType' (Just te') _ | niceTypeExp te' = False
-    nastyType' _ t' = nastyType t'
-
-nastyParameter :: Pat ParamType -> Bool
-nastyParameter p = nastyType (patternType p) && not (ascripted p)
-  where
-    ascripted (PatAscription _ te _) = niceTypeExp te
-    ascripted (PatParens p' _) = ascripted p'
-    ascripted _ = False
-
-niceTypeExp :: TypeExp Exp VName -> Bool
-niceTypeExp (TEVar (QualName [] _) _) = True
-niceTypeExp (TEApply te TypeArgExpSize {} _) = niceTypeExp te
-niceTypeExp (TEArray _ te _) = niceTypeExp te
-niceTypeExp (TEUnique te _) = niceTypeExp te
-niceTypeExp (TEDim _ te _) = niceTypeExp te
-niceTypeExp _ = False
-
 checkOneDec :: DecBase NoInfo Name -> TypeM (TySet, Env, DecBase Info VName)
 checkOneDec (ModDec struct) = do
   (abs, modenv, struct') <- checkModBind struct
