@@ -9,6 +9,7 @@ module Language.Futhark.TypeChecker.Modules
 where
 
 import Control.Monad
+import Control.Monad.Identity
 import Data.Either
 import Data.Map.Strict qualified as M
 import Data.Maybe
@@ -17,6 +18,7 @@ import Data.Set qualified as S
 import Futhark.Util.Pretty
 import Language.Futhark
 import Language.Futhark.Semantic
+import Language.Futhark.Traversals
 import Language.Futhark.TypeChecker.Monad
 import Language.Futhark.TypeChecker.Types
 import Language.Futhark.TypeChecker.Unify (doUnification)
@@ -109,9 +111,20 @@ newNamesForMTy orig_mty = do
         substitute v =
           fromMaybe v $ M.lookup v substs
 
-        -- For applySubst and friends.
-        subst v =
-          ExpSubst . flip sizeFromName mempty . qualName <$> M.lookup v substs
+        substituteInExp :: Exp -> Exp
+        substituteInExp = runIdentity . astMap mapper
+          where
+            mapper =
+              ASTMapper
+                { mapOnExp = pure . substituteInExp,
+                  mapOnName = pure . substituteInQualName,
+                  mapOnStructType = pure . substituteInType,
+                  mapOnParamType = pure . substituteInType,
+                  mapOnResRetType = pure . substituteInRetType
+                }
+
+        substituteInQualName (QualName qs v) =
+          QualName (map substitute qs) (substitute v)
 
         substituteInMap f m =
           let (ks, vs) = unzip $ M.toList m
@@ -143,9 +156,8 @@ newNamesForMTy orig_mty = do
           TypeParamType l (substitute p) loc
 
         substituteInScalarType :: ScalarTypeBase Size u -> ScalarTypeBase Size u
-        substituteInScalarType (TypeVar u (QualName qs v) targs) =
-          TypeVar u (QualName (map substitute qs) $ substitute v) $
-            map substituteInTypeArg targs
+        substituteInScalarType (TypeVar u v targs) =
+          TypeVar u (substituteInQualName v) $ map substituteInTypeArg targs
         substituteInScalarType (Prim t) =
           Prim t
         substituteInScalarType (Record ts) =
@@ -155,15 +167,18 @@ newNamesForMTy orig_mty = do
         substituteInScalarType (Arrow als v d1 t1 (RetType dims t2)) =
           Arrow als v d1 (substituteInType t1) $ RetType dims $ substituteInType t2
 
+        substituteInRetType :: RetTypeBase Size u -> RetTypeBase Size u
+        substituteInRetType (RetType ext t) = RetType ext $ substituteInType t
+
         substituteInType :: TypeBase Size u -> TypeBase Size u
         substituteInType (Scalar t) = Scalar $ substituteInScalarType t
         substituteInType (Array u shape t) =
           Array u (substituteInShape shape) $ substituteInScalarType t
 
-        substituteInShape (Shape ds) = Shape $ map (applySubst subst) ds
+        substituteInShape (Shape ds) = Shape $ map substituteInExp ds
 
         substituteInTypeArg (TypeArgDim e) =
-          TypeArgDim $ applySubst subst e
+          TypeArgDim $ substituteInExp e
         substituteInTypeArg (TypeArgType t) =
           TypeArgType $ substituteInType t
 
@@ -182,13 +197,19 @@ envTypeAbbrs env =
     <> (mconcat . map modTypeAbbrs . M.elems . envModTable) env
 
 -- | Refine the given type name in the given env.
+--
+-- XXX: we do not check whether this results in a meaningful module type. In
+-- particular, we may refine a nonlifted type to contain a function or
+-- existentially quantified sizes. However, it is still not possible to
+-- construct a module that matches such malformed module types, so this is not a
+-- soundness issue, merely an ergonomic issue.
 refineEnv ::
   SrcLoc ->
   TySet ->
   Env ->
   QualName Name ->
   [TypeParam] ->
-  StructType ->
+  StructRetType ->
   TypeM (QualName VName, TySet, Env)
 refineEnv loc tset env tname ps t
   | Just (tname', TypeAbbr _ cur_ps (RetType _ (Scalar (TypeVar _ (QualName qs v) _)))) <-
@@ -202,8 +223,8 @@ refineEnv loc tset env tname ps t
               substituteTypesInEnv
                 ( flip M.lookup $
                     M.fromList
-                      [ (qualLeaf tname', Subst cur_ps $ RetType [] t),
-                        (v, Subst ps $ RetType [] t)
+                      [ (qualLeaf tname', Subst cur_ps t),
+                        (v, Subst ps t)
                       ]
                 )
                 env
@@ -343,12 +364,12 @@ resolveMTyNames = resolveMTyNames'
 missingType :: (Pretty a) => Loc -> a -> Either TypeError b
 missingType loc name =
   Left . TypeError loc mempty $
-    "Module does not define a type named" <+> pretty name <> "."
+    "Module does not define a type named" <+> dquotes (pretty name) <> "."
 
 missingVal :: (Pretty a) => Loc -> a -> Either TypeError b
 missingVal loc name =
   Left . TypeError loc mempty $
-    "Module does not define a value named" <+> pretty name <> "."
+    "Module does not define a value named" <+> dquotes (pretty name) <> "."
 
 topLevelSize :: Loc -> VName -> Either TypeError b
 topLevelSize loc name =
@@ -358,7 +379,7 @@ topLevelSize loc name =
 missingMod :: (Pretty a) => Loc -> a -> Either TypeError b
 missingMod loc name =
   Left . TypeError loc mempty $
-    "Module does not define a module named" <+> pretty name <> "."
+    "Module does not define a module named" <+> dquotes (pretty name) <> "."
 
 mismatchedType ::
   Loc ->
