@@ -5,7 +5,7 @@ set -e
 with_backend() {
     echo
     echo "Trying backend $1"
-    futhark bench --backend=$1 --profile --json prog.json prog.fut
+    $2 futhark bench --backend=$1 --profile --json prog.json prog.fut
     rm -rf prog.prof
     futhark profile prog.json
 }
@@ -13,3 +13,5 @@ with_backend() {
 with_backend c
 
 with_backend multicore
+
+with_backend opencl oclgrind
